/*
--             This file is part of the New World OS project
--                 Copyright (C) 2007-2008  QRW Software
--           J. Scott Edwards - j.scott.edwards.nwos@gmail.com 
--                      http://www.qrwsoftware.com
--                      http://nwos.sourceforge.com
--
--   This program is free software: you can redistribute it and/or modify
--   it under the terms of the GNU General Public License as published by
--   the Free Software Foundation, either version 3 of the License, or
--   (at your option) any later version.
--
--   This program is distributed in the hope that it will be useful,
--   but WITHOUT ANY WARRANTY; without even the implied warranty of
--   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--   GNU General Public License for more details.
--
--   You should have received a copy of the GNU General Public License
--   along with this program, in the file LICENSE.  If not, see 
--   <http://www.gnu.org/licenses/>.
--
--   You can also contact me via paper mail at:
--
--      QRW Software
--      P.O. Box 27511
--      Salt Lake City, UT 84127-0511, USA.
--
--
-- This is a one time use program to add a new revision to the File class.
--
-- $Log: add_class_revision.c,v $
-- Revision 1.7  2008/09/01 00:22:50  jsedwards
-- Fix year in copyright.  NO code changes.
--
-- Revision 1.6  2008/08/31 18:25:41  jsedwards
-- Change for new parameters to nwos_initialize_objectify without key stuff
-- and eliminate calle to nwos_set_root_object because initialize does that now.
--
-- Revision 1.5  2008/08/31 16:33:38  jsedwards
-- Added assert around call to nwos_read_variable_sized_object_from_disk
-- because it now returns false if it fails, whereas before it asserted itself.
--
-- Revision 1.4  2008/05/24 01:52:43  jsedwards
-- Modified to create yet another FILE class revision, this time adding the
-- 'source' feature.
--
-- Revision 1.3  2008/03/14 12:04:39  jsedwards
-- Changed to add another FILE class revision with SHA512 checksums.
--
-- Revision 1.2  2007/07/28 01:01:33  jsedwards
-- Changed to only use 5 bytes for file size (at least for now).
--
-- Revision 1.1  2007/07/27 17:14:41  jsedwards
-- Initial version to add new revision of File class.
--
*/


#include <assert.h>
#include <ctype.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "../crc32.h"
#include "../objectify_private.h"


static size_t get_class_object_size(void* class_obj)
{
    return sizeof(C_struct_Class_Definition) + (((C_struct_Class_Definition*)class_obj)->count * sizeof(ObjRef));
}


ObjRef* byte_class_ref()
{
    static ObjRef result;

    if (is_void_reference(&result))
    {
	if (!nwos_find_public_class_definition("BYTE", &result))
	{
	    fprintf(stderr, "Could not locate BYTE class definition\n");
	    nwos_terminate_objectify();
	    exit(1);
	}

	printf("byte_class_ref: %02x%02x%02x%02x\n",
	       result.id[0],
	       result.id[1],
	       result.id[2],
	       result.id[3]);
    }

    return &result;
}


ObjRef* character_class_ref()
{
    static ObjRef result;

    if (is_void_reference(&result))
    {
	if (!nwos_find_public_class_definition("CHARACTER", &result))
	{
	    fprintf(stderr, "Could not locate CHARACTER class definition\n");
	    nwos_terminate_objectify();
	    exit(1);
	}

	printf("byte_class_ref: %02x%02x%02x%02x\n",
	       result.id[0],
	       result.id[1],
	       result.id[2],
	       result.id[3]);
    }

    return &result;
}


ObjRef* objref_class_ref()
{
    static ObjRef result;

    if (is_void_reference(&result))
    {
	if (!nwos_find_public_class_definition("OBJECT REFERENCE", &result))
	{
	    fprintf(stderr, "Could not locate OBJECT REFERENCE class definition\n");
	    nwos_terminate_objectify();
	    exit(1);
	}

	printf("object_reference_class_ref: %02x%02x%02x%02x\n",
	       result.id[0],
	       result.id[1],
	       result.id[2],
	       result.id[3]);
    }

    return &result;
}


void add_feature(C_struct_Class_Definition* class_obj_ptr, ObjRef* type, char* name, int num_elements)
{
    int count;
    ObjRef feature_ref;

    assert(get_class_object_size(class_obj_ptr) < FILE_BLOCK_SIZE);

    count = class_obj_ptr->count;

    nwos_find_or_create_public_feature_definition(type, name, num_elements, &feature_ref);

    copy_reference(&class_obj_ptr->feature[count], &feature_ref);

    nwos_add_to_references(&class_obj_ptr->header.common.id, &feature_ref);

    class_obj_ptr->count = count + 1;

    printf("feature: %d - %s  new size: %d\n", count, name, (int)get_class_object_size(class_obj_ptr));
}



int main(int argc, char* argv[])
{
    ObjRef old_class_ref;
    ObjRef new_class_ref;
    ObjRef name_ref;
    uint8 kludge[FILE_BLOCK_SIZE];
    C_struct_Class_Definition* class_obj_ptr = (C_struct_Class_Definition*)kludge;


    nwos_log_arguments(argc, argv);

    nwos_initialize_objectify(PUBLIC, NULL);


    nwos_find_public_class_definition("FILE", &old_class_ref);

    printf("class_ref: %02x%02x%02x%02x\n",
	   old_class_ref.id[0],
	   old_class_ref.id[1],
	   old_class_ref.id[2],
	   old_class_ref.id[3]);

    assert(nwos_read_variable_sized_object_from_disk(&old_class_ref, kludge, sizeof(kludge), &get_class_object_size));

    printf("  number of features: %d\n", class_obj_ptr->count);

    if (class_obj_ptr->count != 7)
    {
	fprintf(stderr, "File class doesn't have 6 features\n");
	exit(1);
    }

    /* first modify the old class definition to point to the new class */

    nwos_generate_new_id(&new_class_ref);  /* reference for new File class */

    copy_reference(&class_obj_ptr->header.object.next_version, &new_class_ref);

    nwos_crc32_calculate((uint8*) &class_obj_ptr->header.object,
			 sizeof(ObjectHeader),
			 class_obj_ptr->header.common.header_chksum);

    nwos_overwrite_object_to_disk(&old_class_ref, class_obj_ptr, get_class_object_size(class_obj_ptr));

    copy_reference(&name_ref, &class_obj_ptr->name);  /* save the name pointer */

    /* now create the new class definition */

    memset(kludge, 0, sizeof(kludge));

    nwos_fill_in_common_header(&class_obj_ptr->header.common, &new_class_ref, &nwos_public_class_definition_class_ref);

    copy_reference(&class_obj_ptr->name, &name_ref);   /* use the same name */

    copy_reference(&class_obj_ptr->header.object.prev_version, &old_class_ref);

    nwos_create_reference_list(&new_class_ref, &class_obj_ptr->header.object.references);

    printf("class_ref: %02x%02x%02x%02x\n",
	   new_class_ref.id[0],
	   new_class_ref.id[1],
	   new_class_ref.id[2],
	   new_class_ref.id[3]);

    add_feature(class_obj_ptr, byte_class_ref(),   "size", 5);
    add_feature(class_obj_ptr, objref_class_ref(), "source", 1);
    add_feature(class_obj_ptr, objref_class_ref(), "md5sum", 1);
    add_feature(class_obj_ptr, objref_class_ref(), "sha1sum", 1);
    add_feature(class_obj_ptr, objref_class_ref(), "sha256sum", 1);
    add_feature(class_obj_ptr, objref_class_ref(), "sha512sum", 1);
    add_feature(class_obj_ptr, objref_class_ref(), "media", 1);
    add_feature(class_obj_ptr, objref_class_ref(), "block list", 1);

    nwos_crc32_calculate((uint8*) &class_obj_ptr->header.object,
			 sizeof(ObjectHeader),
			 class_obj_ptr->header.common.header_chksum);

    nwos_crc32_calculate((uint8*) &class_obj_ptr->name,
			 get_class_object_size(class_obj_ptr) - sizeof(EveryObject),
			 class_obj_ptr->header.common.data_chksum);

    nwos_write_object_to_disk(&new_class_ref, class_obj_ptr, get_class_object_size(class_obj_ptr));

    nwos_add_to_references(&new_class_ref, &nwos_public_class_definition_class_ref);
    nwos_add_to_references(&new_class_ref, &class_obj_ptr->name);

    nwos_terminate_objectify();

    return 0;
}

