/*
--             This file is part of the New World OS project
--                 Copyright (C) 2007-2008  QRW Software
--           J. Scott Edwards - j.scott.edwards.nwos@gmail.com 
--                      http://www.qrwsoftware.com
--                      http://nwos.sourceforge.com
--
--   This program is free software: you can redistribute it and/or modify
--   it under the terms of the GNU General Public License as published by
--   the Free Software Foundation, either version 3 of the License, or
--   (at your option) any later version.
--
--   This program is distributed in the hope that it will be useful,
--   but WITHOUT ANY WARRANTY; without even the implied warranty of
--   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--   GNU General Public License for more details.
--
--   You should have received a copy of the GNU General Public License
--   along with this program, in the file LICENSE.  If not, see 
--   <http://www.gnu.org/licenses/>.
--
--   You can also contact me via paper mail at:
--
--      QRW Software
--      P.O. Box 27511
--      Salt Lake City, UT 84127-0511, USA.
--
--
-- $Log: add_credit_union.c,v $
-- Revision 1.10  2008/09/01 00:22:51  jsedwards
-- Fix year in copyright.  NO code changes.
--
-- Revision 1.9  2008/08/31 16:24:44  jsedwards
-- Added assert around call to nwos_read_object_from_disk because now it
-- returns false if it fails, before it asserted itself if it failed.
--
-- Revision 1.8  2007/07/01 19:44:12  jsedwards
-- Upgrade to GPLv3.
--
-- Revision 1.7  2007/05/24 12:47:53  jsedwards
-- Change to not add credit union to acronym list until after it is created.
--
-- Revision 1.6  2007/05/24 12:07:05  jsedwards
-- Fix acronym stuff.
--
-- Revision 1.5  2007/05/24 11:56:35  jsedwards
-- Added acronym entry.
--
-- Revision 1.4  2007/04/22 15:18:53  jsedwards
-- Changed to allow short_name and url_list to be null.
--
-- Revision 1.3  2007/04/22 14:24:38  jsedwards
-- Added code to read back object, test it and print it out.
--
-- Revision 1.2  2007/04/22 14:23:49  jsedwards
-- Added code to read and create URL list.
--
-- Revision 1.1  2007/04/21 10:36:57  jsedwards
-- Initial version.
--
*/

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../objectify_private.h"
#include "../crc32.h"

#define MAX_NAME_SIZE 96
#define MAX_URL_SIZE 64
#define MAX_ACRONYM_SIZE 8


static void get_input(char* descr, char* buffer, size_t size)
{
    char *ptr;

    while (1)
    {
	printf("%s: ", descr);
	fflush(stdout);
	fgets(buffer, size, stdin);
	ptr = strchr(buffer, '\n');
	if (ptr != NULL)
	{
	    *ptr = '\0';
	    break;
	}
	do { fgets(buffer, size, stdin); } while (strchr(buffer, '\n') == NULL);
	printf("input too long - try again!\n");
    }
}

extern ObjCreateResult nwos_create_url(char* url, ObjRef* ref);

int main(int argc, char* argv[])
{
    ObjRef root_ref;
    ObjRef reference;
    char full_name[MAX_NAME_SIZE];
    char short_name[MAX_NAME_SIZE];
    char acronym[MAX_ACRONYM_SIZE];
    char url[MAX_URL_SIZE];
    ObjRef cu_class_ref;
    ObjRef url_ref;
    C_struct_Credit_Union cu_obj;
    C_struct_Credit_Union test_cu_obj;
    char test_string[128];
    ReferenceList* ref_list;
    int num_refs;


    nwos_log_arguments(argc, argv);

    nwos_initialize_objectify(NULL, 0, 0, DEFAULT_TYPE_RW, DEFAULT_FILE);

    nwos_set_encryption_level(Encryption_None);

    printf("Adding credit union on: %s\n", DEFAULT_FILE);

    printf("next_reference: %02x%02x%02x%02x\n", 
	   nwos_next_public_ref.id[0],
	   nwos_next_public_ref.id[1],
	   nwos_next_public_ref.id[2],
	   nwos_next_public_ref.id[3]);

    root_ref.id[0] = 0;
    root_ref.id[1] = 0;
    root_ref.id[2] = 0;
    root_ref.id[3] = 1;

    nwos_set_root_object(&root_ref);


    assert(nwos_find_public_class_definition("CREDIT UNION", &cu_class_ref));


    get_input("full name", full_name, sizeof(full_name));
    get_input("short name", short_name, sizeof(short_name));
    get_input("acronym", acronym, sizeof(acronym));
    get_input("url", url, sizeof(url));


    if (nwos_ask_yes_or_no(NULL, "Create?"))
    {
	memset(&cu_obj, 0, sizeof(cu_obj));

	nwos_generate_new_id(&reference);

	nwos_fill_in_common_header(&cu_obj.header.common, &reference, &cu_class_ref);

	nwos_create_name(full_name, &cu_obj.full_name);

	if (*short_name != '\0')
	{
	    nwos_create_name(short_name, &cu_obj.short_name);
	}

	if (*url != '\0')
	{
	    nwos_create_reference_list(&reference, &cu_obj.url_list);
	    nwos_create_url(url, &url_ref);
	    nwos_add_to_reference_list(&url_ref, &cu_obj.url_list);
	}

	if (*acronym != '\0')
	{
	    nwos_create_public_acronym(acronym, &cu_obj.acronym);
	}

	nwos_create_reference_list(&reference, &cu_obj.header.object.references);

	nwos_crc32_calculate((uint8*) &cu_obj.header.object, sizeof(ObjectHeader), cu_obj.header.common.header_chksum);

	nwos_crc32_calculate((uint8*) &cu_obj.full_name, sizeof(cu_obj) - sizeof(EveryObject), cu_obj.header.common.data_chksum);

	nwos_write_object_to_disk(&reference, &cu_obj, sizeof(cu_obj));

        nwos_add_to_references(&reference, &cu_class_ref);
	nwos_add_to_references(&reference, &cu_obj.full_name);

	if (!is_void_reference(&cu_obj.short_name))
	{
	    nwos_add_to_references(&reference, &cu_obj.short_name);
	}

	if (!is_void_reference(&cu_obj.url_list))
	{
	    nwos_add_to_references(&reference, &url_ref);
	}

	if (!is_void_reference(&cu_obj.acronym))
	{
	    nwos_add_to_references(&reference, &cu_obj.acronym);
	}

	printf("Object created: %02x%02x%02x%02x\n", reference.id[0], reference.id[1], reference.id[2], reference.id[3]);
    }

    assert(nwos_read_object_from_disk(&reference, &test_cu_obj, sizeof(test_cu_obj)));

    if (memcmp(&test_cu_obj, &cu_obj, sizeof(cu_obj)) != 0)
    {
	printf("Error: object read back doesn't match original object\n");
    }
    else
    {
	assert(nwos_name_to_string(&test_cu_obj.full_name, test_string, sizeof(test_string)));
	printf("full name: %s\n", test_string);

	if (!is_void_reference(&test_cu_obj.short_name))
	{
	    assert(nwos_name_to_string(&test_cu_obj.short_name, test_string, sizeof(test_string)));
	    printf("short name: %s\n", test_string);
	}

	if (!is_void_reference(&test_cu_obj.url_list))
	{
	    ref_list = nwos_malloc_reference_list(&test_cu_obj.url_list);

	    num_refs = ref_list->common_header.num_refs;

	    if (num_refs != 1)
	    {
		printf("Error: number of URLs is not 1: %d\n", num_refs);
	    }
	    else
	    {
		assert(nwos_url_to_string(&ref_list->references[0], test_string, sizeof(test_string)));
		printf("url: %s\n", test_string);
	    }

	    nwos_free_reference_list(ref_list);
	    ref_list = NULL;
	}
    }

    nwos_terminate_objectify();

    return 0;
}


