/*
--             This file is part of the New World OS project
--                 Copyright (C) 2006-2007  QRW Software
--           J. Scott Edwards - j.scott.edwards.nwos@gmail.com 
--                      http://www.qrwsoftware.com
--                      http://nwos.sourceforge.com
--
--   This program is free software: you can redistribute it and/or modify
--   it under the terms of the GNU General Public License as published by
--   the Free Software Foundation, either version 3 of the License, or
--   (at your option) any later version.
--
--   This program is distributed in the hope that it will be useful,
--   but WITHOUT ANY WARRANTY; without even the implied warranty of
--   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--   GNU General Public License for more details.
--
--   You should have received a copy of the GNU General Public License
--   along with this program, in the file LICENSE.  If not, see 
--   <http://www.gnu.org/licenses/>.
--
--   You can also contact me via paper mail at:
--
--      QRW Software
--      P.O. Box 27511
--      Salt Lake City, UT 84127-0511, USA.
--
--
-- $Log: import_from_0005_csv.c,v $
-- Revision 1.2  2007/07/01 19:44:12  jsedwards
-- Upgrade to GPLv3.
--
-- Revision 1.1  2006/12/02 17:37:03  jsedwards
-- Moved from root directory because likely won't be used again.
--
-- Revision 1.4  2006/11/11 12:01:04  jsedwards
-- Update e-mail address to something that works.
--
-- Revision 1.3  2006/11/03 04:17:12  jsedwards
-- Change "nick_name" to "nickname" (in the dictionary it is one word).
--
-- Revision 1.2  2006/10/26 01:51:27  jsedwards
-- Merged alpha_05_branch back into main trunk.
--
-- Revision 1.1.2.11  2006/10/25 12:22:28  jsedwards
-- Changed C_struct_class_definition to C_struct_Class_Definition so the case
-- is consistent with all the other C_struct objects.
--
-- Revision 1.1.2.10  2006/10/24 03:23:38  jsedwards
-- Changed assert to check that number of public ids matches instead of
-- number reserved public blocks (reserved public blocks is variable now).
--
-- Revision 1.1.2.9  2006/10/03 12:53:07  jsedwards
-- Changed so that instead of calling a separate routine after initialize to
-- change the already opened storage, you call it now with a type of storage
-- parameter and a path to the storage.  The problem with the other way was
-- if you tried reading a compressed file on another machine it tried to open
-- the default file which didn't exist.
--
-- Revision 1.1.2.8  2006/09/29 13:07:01  jsedwards
-- Fixed bug in if statements where only first update_objref that returned
-- true was called (so multiple ones wouldn't be translated in object).
--
-- Revision 1.1.2.7  2006/09/29 04:21:58  jsedwards
-- Changed to use magic number and version string definitions instead of
-- hard coding the values.
--
-- Revision 1.1.2.6  2006/09/29 04:09:14  jsedwards
-- Changed to convert all reference id's that have a zero in the uppermost
-- nibble to one that doesn't.
--
-- Revision 1.1.2.5  2006/09/27 12:36:18  jsedwards
-- Moved checking of object header checksum to import_every_object routine
-- (this saved having code in every import_xxx routine to do it).
--
-- Revision 1.1.2.4  2006/09/26 13:27:40  jsedwards
-- Added code to make sure file was opened ok.
--
-- Revision 1.1.2.3  2006/09/17 14:53:20  jsedwards
-- Removed left over code from export file.
--
-- Revision 1.1.2.2  2006/09/17 12:20:11  jsedwards
-- Added call to terminate_objectify() so all reference lists get flushed to
-- the file.
--
-- Revision 1.1.2.1  2006/09/16 14:22:40  jsedwards
-- Initial version, can read csv files exported by export_0005_to_csv.
--
*/

#include <assert.h>
#include <ctype.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "crc32.h"
#include "objectify_private.h"


char* current_filename;   /* kludgy, but needed for printing */

FILE* ref_list_fp;

ObjRef nwos_root_object_ref;


#define MAX_LINE_LENGTH 256
#define MAX_REF_LIST_LINE_LENGTH 8192


static size_t get_spelling_object_size(void* spelling_obj)
{
    assert(((C_struct_Spelling*)spelling_obj)->count > 0);
    return sizeof(C_struct_Spelling) + ((C_struct_Spelling*)spelling_obj)->count;
}

static size_t get_name_object_size(void* name_obj)
{
    assert(((C_struct_Name*)name_obj)->count > 0);
    return sizeof(C_struct_Name) + (((C_struct_Name*)name_obj)->count * sizeof(ObjRef));
}

static size_t get_year_object_size(void* year_obj)
{
    assert(((C_struct_Year*)year_obj)->count > 0);
    return sizeof(C_struct_Year) + ((C_struct_Year*)year_obj)->count;
}


void shift_line(char* line, int index)
{
    int i;

    assert(strlen(line) < (MAX_REF_LIST_LINE_LENGTH - index));

    for (i = 0; line[i + index] != '\0'; i++)
    {
	line[i] = line[i + index];
    }

    line[i] = '\0';
}


void fput_char(char c, FILE* fp)
{
    if (fputc(c, fp) == EOF)
    {
	perror(current_filename);
	exit(1);
    }
}


void check_comma(char line[MAX_LINE_LENGTH])
{
    if (line[0] != ',')
    {
	printf("%s: comma\n", current_filename);
	printf("remainder of line: %s\n", line);
	exit(1);
    }

    shift_line(line, 1);
}


void check_double_quote(char line[MAX_LINE_LENGTH])
{
    if (line[0] != '"')
    {
	printf("%s: missing double quote\n", current_filename);
	printf("remainder of line: %s\n", line);
	exit(1);
    }

    shift_line(line, 1);
}


void check_newline(char line[MAX_LINE_LENGTH])
{
    if (line[0] != '\n')
    {
	printf("%s: missing end of line\n", current_filename);
	printf("remainder of line: %s\n", line);
	exit(1);
    }

    shift_line(line, 1);
}

void get_string(char line[MAX_LINE_LENGTH], char string[], int size)
{
    int i;

    check_double_quote(line);

    for (i = 0; line[i] != '"' && line[i] != '\0'; i++)
    {
	if (i >= (size - 1))
	{
	    printf("%s: string too long\n", current_filename);
	    printf("remainder of line: %s\n", line);
	    exit(1);
	}

	string[i] = line[i];
    }

    string[i] = '\0';

    shift_line(line, i);

    check_double_quote(line);
}


void get_uint8(char line[MAX_LINE_LENGTH], uint8* ptr_uint8)
{
    uint8 digit;

    if (!isdigit(line[0]))
    {
	printf("no digit for reading uint8\n");
	printf("remainder of line: %s\n", line);
	exit(1);
    }

    *ptr_uint8 = line[0] - '0';

    shift_line(line, 1);

    if (isdigit(line[0]))
    {
	digit = line[0] - '0';

	shift_line(line, 1);

	*ptr_uint8 = *ptr_uint8 * 10 + digit;
    }

    if (isdigit(line[0]))
    {
	digit = line[0] - '0';

	shift_line(line, 1);

	if (*ptr_uint8 > 25 || (*ptr_uint8 == 25 && line[0] > '6'))
	{
	    printf("value too large for uint8: %u%u\n", *ptr_uint8, digit);
	    printf("remainder of line: %s\n", line);
	    exit(1);
	}

	*ptr_uint8 = *ptr_uint8 * 10 + digit;
    }

    if (isdigit(line[0]))
    {
	printf("too many digits for uint8: %u\n", *ptr_uint8);
	printf("remainder of line: %s\n", line);
	exit(1);
    }
}


int hex_char_to_value(char line[MAX_LINE_LENGTH])
{
    char hex_digit[17] = "0123456789abcdef";
    char* p = strchr(hex_digit, line[0]);

    if (p == NULL)
    {
	printf("%s: invalid hexadecimal character\n", current_filename);
	printf("remainder of line: %s\n", line);
	exit(1);
    }

    shift_line(line, 1);

    return p - hex_digit;
}
    

void get_hex4(char line[MAX_LINE_LENGTH], uint8 hex[4])
{
    int i;
    uint8 upper;
    uint8 lower;

    if (line[0] != '0' || line[1] != 'x')
    {
	printf("%s: missing 0x for hexadecimal value\n", current_filename);
	printf("remainder of line: %s\n", line);
	exit(1);
    }

    shift_line(line, 2);

    for (i = 0; i < 4; i++)
    {
	upper = hex_char_to_value(line);
	lower = hex_char_to_value(line);
	hex[i] = (upper << 4) | lower;
    }
}


void get_objref(char line[MAX_LINE_LENGTH], ObjRef* ref)
{
    get_hex4(line, ref->id);
}


void get_timestamp(char line[MAX_LINE_LENGTH], TimeStamp ts)
{
    char string[32];
    uint year;
    uint month;
    uint day_of_month;
    uint day_of_week;
    uint hour;
    uint minute;
    uint second;
    uint microsecond;
    int count;

    get_string(line, string, sizeof(string));


    count = sscanf(string, "%04u-%02u-%02u (%u) %02u:%02u:%02u.%06u",
		   &year, &month, &day_of_month,  &day_of_week,  &hour, &minute, &second, &microsecond);

    if (count != 8)
    {
	printf("count mismatch when reading time stamp: %s - %d\n", string, count);
	exit(1);
    }

    nwos_insert_year_into_time_stamp(year, ts);
    nwos_insert_month_into_time_stamp(month, ts);
    nwos_insert_day_of_month_into_time_stamp(day_of_month, ts);

    nwos_insert_day_of_week_into_time_stamp(day_of_week, ts);

    nwos_insert_hour_into_time_stamp(hour, ts);
    nwos_insert_minute_into_time_stamp(minute, ts);
    nwos_insert_second_into_time_stamp(second, ts);
    nwos_insert_microsecond_into_time_stamp(microsecond, ts);
}


/* this routine updates a reference that encroaches upon the public space (zero in the uppermost nibble) */
/* objects versions 0004, 0005, and 0011 allowed any 32 bit value except 00000000 */

#define MAX_XLAT 4096

static struct {
    ObjRef old_id;
    ObjRef new_id;
} xlat_old_new_tbl[MAX_XLAT];

bool update_objref(ObjRef* ref)
{
    int i;
    int count = 0;

    assert(NUM_PUBLIC_IDS == 0x10000000);   /* make sure that the upper nibble is matches what is reserved */
                                                    /* if this assert fails, need to fix the while loop below */

    if ((ref->id[0] & 0xF0) == 0 && !is_void_reference(ref))     /* it needs to be changed */
    {
	for (i = 0; !is_void_reference(&xlat_old_new_tbl[i].old_id); i++)     /* find out if it has already been encountered */
	{
	    if (is_same_object(ref, &xlat_old_new_tbl[i].old_id))
	    {
		break;
	    }
	}

	if (is_void_reference(&xlat_old_new_tbl[i].old_id))    /* first time for this one */
	{
	    assert(i < MAX_XLAT);    /* not very graceful */

	    memcpy(&xlat_old_new_tbl[i].old_id, ref, sizeof(ObjRef));    /* save the old one in the table */

	    while (1)
	    {

		xlat_old_new_tbl[i].new_id.word = random();

		if ((xlat_old_new_tbl[i].new_id.id[0] & 0xF0) != 0 && !nwos_block_used(&xlat_old_new_tbl[i].new_id))
		{
		    break;    /* found a good one */
		}

		assert(count++ < 100);    /* don't hang forever if random() is broken */
	    }

	    printf("Change %02x%02x%02x%02x to %02x%02x%02x%02x\n",
		   xlat_old_new_tbl[i].old_id.id[0],
		   xlat_old_new_tbl[i].old_id.id[1],
		   xlat_old_new_tbl[i].old_id.id[2],
		   xlat_old_new_tbl[i].old_id.id[3],
		   xlat_old_new_tbl[i].new_id.id[0],
		   xlat_old_new_tbl[i].new_id.id[1],
		   xlat_old_new_tbl[i].new_id.id[2],
		   xlat_old_new_tbl[i].new_id.id[3]);
	}

	memcpy(ref, &xlat_old_new_tbl[i].new_id, sizeof(ObjRef));

	return true;
    }

    return false;    /* no change needed */
}


void import_common_header(char line[MAX_LINE_LENGTH], CommonHeader* ch)
{
    if (line[0] != '0' || line[1] != '0' || line[2] != '0' || (line[3] != '4' && line[3] != '5'))
    {
	printf("expected object version 0004 or 0005\n");
	printf("remaining line: %s\n", line);
	exit(1);
    }

    shift_line(line, 4);

    check_comma(line);

    get_hex4(line, ch->header_chksum);

    check_comma(line);

    get_hex4(line, ch->data_chksum);

    check_comma(line);

    get_timestamp(line, ch->creation_time);

    check_comma(line);

    get_objref(line, &ch->class_definition);
}


void import_object_header(char line[MAX_LINE_LENGTH], ObjectHeader* oh)
{
    get_objref(line, &oh->references);
}


void import_every_object(char line[MAX_LINE_LENGTH], EveryObject* eo)
{
    uint8 chksum[4];

    get_objref(line, &eo->common.id);

    update_objref(&eo->common.id);    /* can just update this now because it isn't part of a checksum */

    check_comma(line);

    import_common_header(line, &eo->common);

    check_comma(line);

    import_object_header(line, &eo->object);

    check_comma(line);

    nwos_crc32_calculate((uint8*) &eo->object, sizeof(eo->object), chksum);

    assert(memcmp(chksum, eo->common.header_chksum, sizeof(chksum)) == 0);

    if (update_objref(&eo->object.references))   /* it had to be changed, so recompute header checksum */
    {
	nwos_crc32_calculate((uint8*) &eo->object, sizeof(eo->object), eo->common.header_chksum);
    }
}
    

void import_root(char line[MAX_LINE_LENGTH])
{
    C_struct_Root new_root_obj;
    uint8 chksum[4];
    struct {
        EveryObject header;
        ObjRef      class_definition_class;
    } old_root_obj;

    assert(strlen(line) < MAX_LINE_LENGTH);

    memset(&old_root_obj, 0, sizeof(old_root_obj));

    import_every_object(line, &old_root_obj.header);

    printf("Old Root Object %02x%02x%02x%02x: ", 
	   old_root_obj.header.common.id.id[0],
	   old_root_obj.header.common.id.id[1],
	   old_root_obj.header.common.id.id[2],
	   old_root_obj.header.common.id.id[3]);

    get_objref(line, &old_root_obj.class_definition_class);

    check_newline(line);

    printf("class_definition_class: %02x%02x%02x%02x\n",
	   old_root_obj.class_definition_class.id[0],
	   old_root_obj.class_definition_class.id[1],
	   old_root_obj.class_definition_class.id[2],
	   old_root_obj.class_definition_class.id[3]);

    nwos_crc32_calculate((uint8*) &old_root_obj.class_definition_class, sizeof(old_root_obj) - sizeof(EveryObject), chksum);

    assert(memcmp(chksum, old_root_obj.header.common.data_chksum, sizeof(chksum)) == 0);

    memcpy(&nwos_class_definition_class_ref, &old_root_obj.class_definition_class, sizeof(ObjRef));

    assert(!is_void_reference(&nwos_reference_list_class_ref));
    assert(!is_void_reference(&nwos_root_object_ref));

    nwos_create_root(&nwos_root_object_ref);

    printf("New Root Object %02x%02x%02x%02x: ", 
	   new_root_obj.header.common.id.id[0],
	   new_root_obj.header.common.id.id[1],
	   new_root_obj.header.common.id.id[2],
	   new_root_obj.header.common.id.id[3]);
    fflush(stdout);

    nwos_read_object_from_disk(&nwos_root_object_ref, &new_root_obj, sizeof(new_root_obj));

    printf("class_definition_class: %02x%02x%02x%02x  reference_list_class: %02x%02x%02x%02x\n",
	   new_root_obj.class_definition_class.id[0],
	   new_root_obj.class_definition_class.id[1],
	   new_root_obj.class_definition_class.id[2],
	   new_root_obj.class_definition_class.id[3],
	   new_root_obj.reference_list_class.id[0],
	   new_root_obj.reference_list_class.id[1],
	   new_root_obj.reference_list_class.id[2],
	   new_root_obj.reference_list_class.id[3]);
}


void import_class_definition(char line[MAX_LINE_LENGTH])
{
    C_struct_Class_Definition class_def_obj;
    uint8 chksum[4];

    assert(strlen(line) < MAX_LINE_LENGTH);

    memset(&class_def_obj, 0, sizeof(class_def_obj));

    import_every_object(line, &class_def_obj.header);

    printf("Class Def %02x%02x%02x%02x: ", 
	   class_def_obj.header.common.id.id[0],
	   class_def_obj.header.common.id.id[1],
	   class_def_obj.header.common.id.id[2],
	   class_def_obj.header.common.id.id[3]);

    get_objref(line, &class_def_obj.name);

    check_newline(line);

    printf("%02x%02x%02x%02x\n",
	   class_def_obj.name.id[0],
	   class_def_obj.name.id[1],
	   class_def_obj.name.id[2],
	   class_def_obj.name.id[3]);

    nwos_crc32_calculate((uint8*) &class_def_obj.name, sizeof(class_def_obj) - sizeof(EveryObject), chksum);

    assert(memcmp(chksum, class_def_obj.header.common.data_chksum, sizeof(chksum)) == 0);

    if (update_objref(&class_def_obj.name))      /* if name reference changed, recompute data checksum */
    {
	nwos_crc32_calculate((uint8*) &class_def_obj.name,
			     sizeof(class_def_obj) - sizeof(EveryObject), 
			     class_def_obj.header.common.data_chksum);
    }

    nwos_write_object_to_disk(&class_def_obj.header.common.id, &class_def_obj, sizeof(class_def_obj));

    /* kludgy thing to find reference list class */
    if (is_void_reference(&class_def_obj.header.object.references))   /* this must be the reference list class */
    {
	assert(is_void_reference(&nwos_reference_list_class_ref));
	memcpy(&nwos_reference_list_class_ref, &class_def_obj.header.common.id, sizeof(ObjRef));
    }
}


void import_spelling(char line[MAX_LINE_LENGTH])
{
    uint8 kludge[FILE_BLOCK_SIZE];
    C_struct_Spelling* spelling_obj_ptr = (C_struct_Spelling*) kludge;
    int i;
    uint8 chksum[4];

    assert(strlen(line) < MAX_LINE_LENGTH);

    memset(kludge, 0, sizeof(kludge));

    import_every_object(line, &spelling_obj_ptr->header);

    printf("Spelling %02x%02x%02x%02x: ", 
	   spelling_obj_ptr->header.common.id.id[0],
	   spelling_obj_ptr->header.common.id.id[1],
	   spelling_obj_ptr->header.common.id.id[2],
	   spelling_obj_ptr->header.common.id.id[3]);

    check_double_quote(line);

    for (i = 0; line[i] != '"' && line[i] != '\0'; i++)
    {
	assert(i < 100);                          /* there shouldn't be any words this long */
	spelling_obj_ptr->storage[i] = line[i];
	printf("%c", spelling_obj_ptr->storage[i]);
    }

    shift_line(line, i);

    spelling_obj_ptr->count = i;

    check_double_quote(line);

    check_newline(line);

    printf("\n");

    nwos_crc32_calculate((uint8*) &spelling_obj_ptr->character_set, 
			 get_spelling_object_size(spelling_obj_ptr) - sizeof(EveryObject), 
			 chksum);

    assert(memcmp(chksum, spelling_obj_ptr->header.common.data_chksum, sizeof(chksum)) == 0);

    nwos_write_object_to_disk(&spelling_obj_ptr->header.common.id, spelling_obj_ptr, get_spelling_object_size(spelling_obj_ptr));
}
    

void import_name(char line[MAX_LINE_LENGTH])
{
    uint8 kludge[FILE_BLOCK_SIZE];
    C_struct_Name* name_obj_ptr = (C_struct_Name*) kludge;
    int i;
    uint8 chksum[4];
    bool changed;

    assert(strlen(line) < MAX_LINE_LENGTH);

    memset(kludge, 0, sizeof(kludge));

    import_every_object(line, &name_obj_ptr->header);

    printf("Name %02x%02x%02x%02x:", 
	   name_obj_ptr->header.common.id.id[0],
	   name_obj_ptr->header.common.id.id[1],
	   name_obj_ptr->header.common.id.id[2],
	   name_obj_ptr->header.common.id.id[3]);

    i = 0;
    do
    {
	if (i > 0) check_comma(line);

	get_objref(line, &name_obj_ptr->spelling[i]);

	printf(" %02x%02x%02x%02x", 
	   name_obj_ptr->spelling[i].id[0],
	   name_obj_ptr->spelling[i].id[1],
	   name_obj_ptr->spelling[i].id[2],
	   name_obj_ptr->spelling[i].id[3]);

	i++;
    }
    while (line[0] == ',');

    name_obj_ptr->count = i;

    check_newline(line);

    printf("\n");

    nwos_crc32_calculate((uint8*) &name_obj_ptr->count, get_name_object_size(name_obj_ptr) - sizeof(EveryObject), chksum);

    assert(memcmp(chksum, name_obj_ptr->header.common.data_chksum, sizeof(chksum)) == 0);

    changed = false;
    
    for (i = 0; i < name_obj_ptr->count; i++)    /* go back and check all references are legal */
    {
	changed |= update_objref(&name_obj_ptr->spelling[i]);
    }

    if (changed)      /* if any of the spelling references changed, recompute data checksum */
    {
	nwos_crc32_calculate((uint8*) &name_obj_ptr->count,
			     get_name_object_size(name_obj_ptr) - sizeof(EveryObject), 
			     name_obj_ptr->header.common.data_chksum);
    }

    nwos_write_object_to_disk(&name_obj_ptr->header.common.id, name_obj_ptr, get_name_object_size(name_obj_ptr));
}
    

void import_word(char line[MAX_LINE_LENGTH])
{
    C_struct_Word word_obj;
    uint8 chksum[4];

    assert(strlen(line) < MAX_LINE_LENGTH);

    memset(&word_obj, 0, sizeof(word_obj));

    import_every_object(line, &word_obj.header);

    printf("Word %02x%02x%02x%02x: ", 
	   word_obj.header.common.id.id[0],
	   word_obj.header.common.id.id[1],
	   word_obj.header.common.id.id[2],
	   word_obj.header.common.id.id[3]);

    get_objref(line, &word_obj.thing);

    check_comma(line);

    get_objref(line, &word_obj.language);

    check_comma(line);

    get_objref(line, &word_obj.spelling);

    check_newline(line);

    printf("thing: %02x%02x%02x%02x  language: %02x%02x%02x%02x  spelling: %02x%02x%02x%02x\n",
	   word_obj.thing.id[0],
	   word_obj.thing.id[1],
	   word_obj.thing.id[2],
	   word_obj.thing.id[3],
	   word_obj.language.id[0],
	   word_obj.language.id[1],
	   word_obj.language.id[2],
	   word_obj.language.id[3],
	   word_obj.spelling.id[0],
	   word_obj.spelling.id[1],
	   word_obj.spelling.id[2],
	   word_obj.spelling.id[3]);

    nwos_crc32_calculate((uint8*) &word_obj.thing, sizeof(word_obj) - sizeof(EveryObject), chksum);

    assert(memcmp(chksum, word_obj.header.common.data_chksum, sizeof(chksum)) == 0);

    /* if any of the references changed, recompute the data checksum */
    if (update_objref(&word_obj.thing) | update_objref(&word_obj.language) | update_objref(&word_obj.spelling))
    {
	nwos_crc32_calculate((uint8*) &word_obj.thing,
			     sizeof(word_obj) - sizeof(EveryObject),
			     word_obj.header.common.data_chksum);
    }

    nwos_write_object_to_disk(&word_obj.header.common.id, &word_obj, sizeof(word_obj));
}
    

void import_language(char line[MAX_LINE_LENGTH])
{
    C_struct_Language language_obj;
    uint8 chksum[4];

    assert(strlen(line) < MAX_LINE_LENGTH);

    memset(&language_obj, 0, sizeof(language_obj));

    import_every_object(line, &language_obj.header);

    printf("Language %02x%02x%02x%02x: ", 
	   language_obj.header.common.id.id[0],
	   language_obj.header.common.id.id[1],
	   language_obj.header.common.id.id[2],
	   language_obj.header.common.id.id[3]);

    get_objref(line, &language_obj.definition);

    check_newline(line);

    printf("%02x%02x%02x%02x\n",
	   language_obj.definition.id[0],
	   language_obj.definition.id[1],
	   language_obj.definition.id[2],
	   language_obj.definition.id[3]);

    nwos_crc32_calculate((uint8*) &language_obj.definition, sizeof(language_obj) - sizeof(EveryObject), chksum);

    assert(memcmp(chksum, language_obj.header.common.data_chksum, sizeof(chksum)) == 0);

    if (update_objref(&language_obj.definition))
    {
	nwos_crc32_calculate((uint8*) &language_obj.definition,
			     sizeof(language_obj) - sizeof(EveryObject),
			     language_obj.header.common.data_chksum);
    }

    nwos_write_object_to_disk(&language_obj.header.common.id, &language_obj, sizeof(language_obj));
}
    

void import_us_state(char line[MAX_LINE_LENGTH])
{
    C_struct_US_State state_obj;
    uint8 chksum[4];

    assert(strlen(line) < MAX_LINE_LENGTH);

    memset(&state_obj, 0, sizeof(state_obj));

    import_every_object(line, &state_obj.header);

    printf("State %02x%02x%02x%02x: ", 
	   state_obj.header.common.id.id[0],
	   state_obj.header.common.id.id[1],
	   state_obj.header.common.id.id[2],
	   state_obj.header.common.id.id[3]);

    get_objref(line, &state_obj.name);

    check_comma(line);

    get_objref(line, &state_obj.date);

    check_comma(line);

    get_objref(line, &state_obj.capital);

    check_comma(line);

    get_objref(line, &state_obj.postal_code);

    check_comma(line);

    get_uint8(line, &state_obj.number);

    check_newline(line);

    printf("name: %02x%02x%02x%02x  date: %02x%02x%02x%02x  capital: %02x%02x%02x%02x  abbr: %02x%02x%02x%02x  num:%d\n",
	   state_obj.name.id[0], state_obj.name.id[1], state_obj.name.id[2], state_obj.name.id[3],
	   state_obj.date.id[0], state_obj.date.id[1], state_obj.date.id[2], state_obj.date.id[3],
	   state_obj.capital.id[0], state_obj.capital.id[1], state_obj.capital.id[2], state_obj.capital.id[3],
	   state_obj.postal_code.id[0], state_obj.postal_code.id[1], state_obj.postal_code.id[2], state_obj.postal_code.id[3],
	   state_obj.number);

    nwos_crc32_calculate((uint8*) &state_obj.name, sizeof(state_obj) - sizeof(EveryObject), chksum);

    assert(memcmp(chksum, state_obj.header.common.data_chksum, sizeof(chksum)) == 0);

    if (update_objref(&state_obj.name) | update_objref(&state_obj.date) |
	update_objref(&state_obj.capital) | update_objref(&state_obj.postal_code))
    {
	nwos_crc32_calculate((uint8*) &state_obj.name,
			     sizeof(state_obj) - sizeof(EveryObject),
			     state_obj.header.common.data_chksum);
    }

    nwos_write_object_to_disk(&state_obj.header.common.id, &state_obj, sizeof(state_obj));
}
    

void import_us_city(char line[MAX_LINE_LENGTH])
{
    C_struct_US_City city_obj;
    uint8 chksum[4];

    assert(strlen(line) < MAX_LINE_LENGTH);

    memset(&city_obj, 0, sizeof(city_obj));

    import_every_object(line, &city_obj.header);

    printf("City %02x%02x%02x%02x: ", 
	   city_obj.header.common.id.id[0],
	   city_obj.header.common.id.id[1],
	   city_obj.header.common.id.id[2],
	   city_obj.header.common.id.id[3]);

    get_objref(line, &city_obj.name);

    check_comma(line);

    get_objref(line, &city_obj.state);

    check_newline(line);

    printf("name: %02x%02x%02x%02x  state: %02x%02x%02x%02x\n",
	   city_obj.name.id[0], city_obj.name.id[1], city_obj.name.id[2], city_obj.name.id[3],
	   city_obj.state.id[0], city_obj.state.id[1], city_obj.state.id[2], city_obj.state.id[3]);

    nwos_crc32_calculate((uint8*) &city_obj.name, sizeof(city_obj) - sizeof(EveryObject), chksum);

    assert(memcmp(chksum, city_obj.header.common.data_chksum, sizeof(chksum)) == 0);

    if (update_objref(&city_obj.name) | update_objref(&city_obj.state))
    {
	nwos_crc32_calculate((uint8*) &city_obj.name, sizeof(city_obj) - sizeof(EveryObject), city_obj.header.common.data_chksum);
    }

    nwos_write_object_to_disk(&city_obj.header.common.id, &city_obj, sizeof(city_obj));
}


void import_area_code(char line[MAX_LINE_LENGTH])
{
    C_struct_Area_Code area_code_obj;
    int i;
    uint8 chksum[4];
    char area_code[4];

    assert(strlen(line) < MAX_LINE_LENGTH);

    memset(&area_code_obj, 0, sizeof(area_code_obj));

    import_every_object(line, &area_code_obj.header);

    printf("Area Code %02x%02x%02x%02x: ", 
	   area_code_obj.header.common.id.id[0],
	   area_code_obj.header.common.id.id[1],
	   area_code_obj.header.common.id.id[2],
	   area_code_obj.header.common.id.id[3]);

    get_objref(line, &area_code_obj.state);

    check_comma(line);

    get_string(line, area_code, sizeof(area_code));

    if (strlen(area_code) != sizeof(area_code_obj.storage))
    {
	printf("area code too short: %s\n", area_code);
	exit(0);
    }

    for (i = 0; i < sizeof(area_code_obj.storage); i++) 
    {
	if (!isdigit(area_code[i]))
	{
	    printf("invalid character in area code: %s\n", area_code);
	    exit(0);
	}
	area_code_obj.storage[i] = area_code[i];
    }

    check_newline(line);

    printf("state: %02x%02x%02x%02x  storage: %c%c%c\n",
	   area_code_obj.state.id[0], area_code_obj.state.id[1], area_code_obj.state.id[2], area_code_obj.state.id[3],
	   area_code_obj.storage[0], area_code_obj.storage[1], area_code_obj.storage[2]);

    nwos_crc32_calculate((uint8*) &area_code_obj.state, sizeof(area_code_obj) - sizeof(EveryObject), chksum);

    assert(memcmp(chksum, area_code_obj.header.common.data_chksum, sizeof(chksum)) == 0);

    if (update_objref(&area_code_obj.state))      /* if state reference changed, update data checksum */
    {
	nwos_crc32_calculate((uint8*) &area_code_obj.state,
			     sizeof(area_code_obj) - sizeof(EveryObject),
			     area_code_obj.header.common.data_chksum);
    }

    nwos_write_object_to_disk(&area_code_obj.header.common.id, &area_code_obj, sizeof(area_code_obj));
}


void import_phone_number(char line[MAX_LINE_LENGTH])
{
    C_struct_Phone_Number phone_num_obj;
    int i;
    uint8 chksum[4];
    char number[8];

    assert(strlen(line) < MAX_LINE_LENGTH);

    memset(&phone_num_obj, 0, sizeof(phone_num_obj));

    import_every_object(line, &phone_num_obj.header);

    printf("Phone Number %02x%02x%02x%02x: ", 
	   phone_num_obj.header.common.id.id[0],
	   phone_num_obj.header.common.id.id[1],
	   phone_num_obj.header.common.id.id[2],
	   phone_num_obj.header.common.id.id[3]);

    get_objref(line, &phone_num_obj.area_code);

    check_comma(line);

    get_string(line, number, sizeof(number));

    if (strlen(number) != sizeof(phone_num_obj.storage))
    {
	printf("area code too short: %s\n", number);
	exit(0);
    }

    for (i = 0; i < sizeof(phone_num_obj.storage); i++) 
    {
	if (!isdigit(number[i]))
	{
	    printf("invalid character in area code: %s\n", number);
	    exit(0);
	}
	phone_num_obj.storage[i] = number[i];
    }

    check_newline(line);

    printf("area_code: %02x%02x%02x%02x  storage: %c%c%c%c%c%c%c\n",
	   phone_num_obj.area_code.id[0], phone_num_obj.area_code.id[1], 
	   phone_num_obj.area_code.id[2], phone_num_obj.area_code.id[3],
	   phone_num_obj.storage[0], phone_num_obj.storage[1], phone_num_obj.storage[2],
	   phone_num_obj.storage[3], phone_num_obj.storage[4], phone_num_obj.storage[5], phone_num_obj.storage[6]);

    nwos_crc32_calculate((uint8*) &phone_num_obj.country, sizeof(phone_num_obj) - sizeof(EveryObject), chksum);

    assert(memcmp(chksum, phone_num_obj.header.common.data_chksum, sizeof(chksum)) == 0);

    if (update_objref(&phone_num_obj.area_code))   /* if phone number reference changed, update data checksum */
    {
	nwos_crc32_calculate((uint8*) &phone_num_obj.country,
			     sizeof(phone_num_obj) - sizeof(EveryObject),
			     phone_num_obj.header.common.data_chksum);
    }

    nwos_write_object_to_disk(&phone_num_obj.header.common.id, &phone_num_obj, sizeof(phone_num_obj));
}


void import_mobile_phone(char line[MAX_LINE_LENGTH])
{
    C_struct_Mobile_Phone mobile_phone_obj;
    uint8 chksum[4];

    assert(strlen(line) < MAX_LINE_LENGTH);

    memset(&mobile_phone_obj, 0, sizeof(mobile_phone_obj));

    import_every_object(line, &mobile_phone_obj.header);

    printf("Mobile Phone %02x%02x%02x%02x: ", 
	   mobile_phone_obj.header.common.id.id[0],
	   mobile_phone_obj.header.common.id.id[1],
	   mobile_phone_obj.header.common.id.id[2],
	   mobile_phone_obj.header.common.id.id[3]);

    get_objref(line, &mobile_phone_obj.person);

    check_comma(line);

    get_objref(line, &mobile_phone_obj.number);

    check_newline(line);

    printf("person: %02x%02x%02x%02x  number: %02x%02x%02x%02x\n",
	   mobile_phone_obj.person.id[0], mobile_phone_obj.person.id[1], 
	   mobile_phone_obj.person.id[2], mobile_phone_obj.person.id[3],
	   mobile_phone_obj.number.id[0], mobile_phone_obj.number.id[1], 
	   mobile_phone_obj.number.id[2], mobile_phone_obj.number.id[3]);

    nwos_crc32_calculate((uint8*) &mobile_phone_obj.person, sizeof(mobile_phone_obj) - sizeof(EveryObject), chksum);

    assert(memcmp(chksum, mobile_phone_obj.header.common.data_chksum, sizeof(chksum)) == 0);

    if (update_objref(&mobile_phone_obj.person) | update_objref(&mobile_phone_obj.number))
    {
	nwos_crc32_calculate((uint8*) &mobile_phone_obj.person,
			     sizeof(mobile_phone_obj) - sizeof(EveryObject),
			     mobile_phone_obj.header.common.data_chksum);
    }

    nwos_write_object_to_disk(&mobile_phone_obj.header.common.id, &mobile_phone_obj, sizeof(mobile_phone_obj));
}


void import_year(char line[MAX_LINE_LENGTH])
{
    uint8 kludge[FILE_BLOCK_SIZE];
    C_struct_Year* year_obj_ptr = (C_struct_Year*) kludge;
    int i;
    uint8 chksum[4];

    assert(strlen(line) < MAX_LINE_LENGTH);

    memset(kludge, 0, sizeof(kludge));

    import_every_object(line, &year_obj_ptr->header);

    printf("Year %02x%02x%02x%02x: ", 
	   year_obj_ptr->header.common.id.id[0],
	   year_obj_ptr->header.common.id.id[1],
	   year_obj_ptr->header.common.id.id[2],
	   year_obj_ptr->header.common.id.id[3]);

    check_double_quote(line);

    for (i = 0; line[i] != '"' && line[i] != '\0'; i++)
    {
	assert(i < 12);                          /* there shouldn't be any ones this long */
	year_obj_ptr->storage[i] = line[i];
	printf("%c", year_obj_ptr->storage[i]);
    }

    shift_line(line, i);

    year_obj_ptr->count = i;

    check_double_quote(line);

    check_newline(line);

    printf("\n");

    nwos_crc32_calculate((uint8*) &year_obj_ptr->count, 
			 get_year_object_size(year_obj_ptr) - sizeof(EveryObject), 
			 chksum);

    assert(memcmp(chksum, year_obj_ptr->header.common.data_chksum, sizeof(chksum)) == 0);

    nwos_write_object_to_disk(&year_obj_ptr->header.common.id, year_obj_ptr, get_year_object_size(year_obj_ptr));
}


void import_month(char line[MAX_LINE_LENGTH])
{
    C_struct_Month month_obj;
    uint8 chksum[4];
    char month[3];
    char* valid_months[12] = { "01", "02", "03", "04", "05", "06", "07", "08", "09", "10", "11", "12" };
    int i;

    assert(strlen(line) < MAX_LINE_LENGTH);

    memset(&month_obj, 0, sizeof(month_obj));

    import_every_object(line, &month_obj.header);

    printf("Month %02x%02x%02x%02x: ", 
	   month_obj.header.common.id.id[0],
	   month_obj.header.common.id.id[1],
	   month_obj.header.common.id.id[2],
	   month_obj.header.common.id.id[3]);

    get_string(line, month, sizeof(month));

    for (i = 0; i < 12; i++) if (strcmp(month, valid_months[i]) == 0) break;

    if (i == 12)
    {
	printf("Invalid month: %s\n", month);
	exit(1);
    }

    month_obj.storage[0] = month[0];
    month_obj.storage[1] = month[1];

    check_comma(line);

    get_uint8(line, &month_obj.minimum_days);

    check_comma(line);

    get_uint8(line, &month_obj.maximum_days);

    check_newline(line);

    printf("month: %c%c  minimum: %u  maximum: %u\n",
	   month_obj.storage[0],
	   month_obj.storage[1],
	   month_obj.minimum_days,
	   month_obj.maximum_days);

    nwos_crc32_calculate((uint8*) &month_obj.definition, sizeof(month_obj) - sizeof(EveryObject), chksum);

    assert(memcmp(chksum, month_obj.header.common.data_chksum, sizeof(chksum)) == 0);

    nwos_write_object_to_disk(&month_obj.header.common.id, &month_obj, sizeof(month_obj));
}
    

void import_date(char line[MAX_LINE_LENGTH])
{
    C_struct_Date date_obj;
    uint8 chksum[4];

    assert(strlen(line) < MAX_LINE_LENGTH);

    memset(&date_obj, 0, sizeof(date_obj));

    import_every_object(line, &date_obj.header);

    printf("Month %02x%02x%02x%02x: ", 
	   date_obj.header.common.id.id[0],
	   date_obj.header.common.id.id[1],
	   date_obj.header.common.id.id[2],
	   date_obj.header.common.id.id[3]);

    get_objref(line, &date_obj.year);

    check_comma(line);

    get_objref(line, &date_obj.month);

    check_comma(line);

    get_uint8(line, &date_obj.day_of_month);

    check_newline(line);

    printf("month: %02x%02x%02x%02x  minimum: %02x%02x%02x%02x  day: %u\n",
	   date_obj.year.id[0], date_obj.year.id[1], date_obj.year.id[2], date_obj.year.id[3],
	   date_obj.month.id[0], date_obj.month.id[1], date_obj.month.id[2], date_obj.month.id[3],
	   date_obj.day_of_month);

    nwos_crc32_calculate((uint8*) &date_obj.year, sizeof(date_obj) - sizeof(EveryObject), chksum);

    assert(memcmp(chksum, date_obj.header.common.data_chksum, sizeof(chksum)) == 0);

    if (update_objref(&date_obj.year) | update_objref(&date_obj.month))
    {
	nwos_crc32_calculate((uint8*) &date_obj.year, sizeof(date_obj) - sizeof(EveryObject), date_obj.header.common.data_chksum);
    }

    nwos_write_object_to_disk(&date_obj.header.common.id, &date_obj, sizeof(date_obj));
}
    

void import_gender(char line[MAX_LINE_LENGTH])
{
    C_struct_Gender gender_obj;
    uint8 chksum[4];

    assert(strlen(line) < MAX_LINE_LENGTH);

    memset(&gender_obj, 0, sizeof(gender_obj));

    import_every_object(line, &gender_obj.header);

    printf("Gender %02x%02x%02x%02x: ", 
	   gender_obj.header.common.id.id[0],
	   gender_obj.header.common.id.id[1],
	   gender_obj.header.common.id.id[2],
	   gender_obj.header.common.id.id[3]);

    get_objref(line, &gender_obj.definition);

    check_newline(line);

    printf("definition: %02x%02x%02x%02x\n",
	   gender_obj.definition.id[0], gender_obj.definition.id[1], gender_obj.definition.id[2], gender_obj.definition.id[3]);

    nwos_crc32_calculate((uint8*) &gender_obj.definition, sizeof(gender_obj) - sizeof(EveryObject), chksum);

    assert(memcmp(chksum, gender_obj.header.common.data_chksum, sizeof(chksum)) == 0);

    if (update_objref(&gender_obj.definition))
    {
	nwos_crc32_calculate((uint8*) &gender_obj.definition,
			     sizeof(gender_obj) - sizeof(EveryObject),
			     gender_obj.header.common.data_chksum);
    }

    nwos_write_object_to_disk(&gender_obj.header.common.id, &gender_obj, sizeof(gender_obj));
}
    

void import_person(char line[MAX_LINE_LENGTH])
{
    C_struct_Person person_obj;
    uint8 chksum[4];
    bool changed;

    assert(strlen(line) < MAX_LINE_LENGTH);

    memset(&person_obj, 0, sizeof(person_obj));

    import_every_object(line, &person_obj.header);

    printf("Person %02x%02x%02x%02x: ", 
	   person_obj.header.common.id.id[0],
	   person_obj.header.common.id.id[1],
	   person_obj.header.common.id.id[2],
	   person_obj.header.common.id.id[3]);

    get_objref(line, &person_obj.gender);

    check_comma(line);

    get_objref(line, &person_obj.first_name);

    check_comma(line);

    get_objref(line, &person_obj.middle_name);

    check_comma(line);

    get_objref(line, &person_obj.last_name);

    check_comma(line);

    get_objref(line, &person_obj.maiden_name);

    check_comma(line);
    
    get_objref(line, &person_obj.goes_by);

    check_comma(line);

    get_objref(line, &person_obj.nickname);

    check_comma(line);

    get_objref(line, &person_obj.birth_date);

    check_newline(line);

    printf("gender: %02x%02x%02x%02x  first:   %02x%02x%02x%02x  middle: %02x%02x%02x%02x  last:  %02x%02x%02x%02x\n",
	   person_obj.gender.id[0], person_obj.gender.id[1], person_obj.gender.id[2], person_obj.gender.id[3],
	   person_obj.first_name.id[0], person_obj.first_name.id[1], person_obj.first_name.id[2], person_obj.first_name.id[3],
	   person_obj.middle_name.id[0], person_obj.middle_name.id[1], 
	   person_obj.middle_name.id[2], person_obj.middle_name.id[3],
	   person_obj.last_name.id[0], person_obj.last_name.id[1], person_obj.last_name.id[2], person_obj.last_name.id[3]);

    printf("                 "); 
    printf("maiden: %02x%02x%02x%02x  goes_by: %02x%02x%02x%02x  nick:   %02x%02x%02x%02x  birth: %02x%02x%02x%02x\n",
	   person_obj.maiden_name.id[0], person_obj.maiden_name.id[1], 
	   person_obj.maiden_name.id[2], person_obj.maiden_name.id[3],
	   person_obj.goes_by.id[0], person_obj.goes_by.id[1], person_obj.goes_by.id[2], person_obj.goes_by.id[3],
	   person_obj.nickname.id[0], person_obj.nickname.id[1], person_obj.nickname.id[2], person_obj.nickname.id[3],
	   person_obj.birth_date.id[0], person_obj.birth_date.id[1], person_obj.birth_date.id[2], person_obj.birth_date.id[3]);

    nwos_crc32_calculate((uint8*) &person_obj.gender, sizeof(person_obj) - sizeof(EveryObject), chksum);

    assert(memcmp(chksum, person_obj.header.common.data_chksum, sizeof(chksum)) == 0);

    changed = update_objref(&person_obj.gender);
    changed |= update_objref(&person_obj.first_name);
    changed |= update_objref(&person_obj.middle_name);
    changed |= update_objref(&person_obj.last_name);
    changed |= update_objref(&person_obj.maiden_name);
    changed |= update_objref(&person_obj.goes_by);
    changed |= update_objref(&person_obj.nickname);
    changed |= update_objref(&person_obj.birth_date);

    if (changed)    // if any of the references changed, recompute the data checksum */
    {
	nwos_crc32_calculate((uint8*) &person_obj.gender,
			     sizeof(person_obj) - sizeof(EveryObject),
			     person_obj.header.common.data_chksum);
    }

    nwos_write_object_to_disk(&person_obj.header.common.id, &person_obj, sizeof(person_obj));
}


void import_reference_list(char *line)
{
    ObjRef ref;
    CommonHeader common_header;
    int num_on_this_line;

    get_objref(line, &common_header.id);

    check_comma(line);

    import_common_header(line, &common_header);

    /* note there may not be a comma after the class definition because the reference list could be empty */

    printf("Reference List %02x%02x%02x%02x:", 
	   common_header.id.id[0],
	   common_header.id.id[1],
	   common_header.id.id[2],
	   common_header.id.id[3]);

    update_objref(&common_header.id);

    nwos_create_reference_list_with_existing_id(NULL, &common_header.id);

    num_on_this_line = 1;

    while (line[0] == ',')
    {
	check_comma(line);

	get_objref(line, &ref);

	update_objref(&ref);

	printf(" %02x%02x%02x%02x", ref.id[0], ref.id[1], ref.id[2], ref.id[3]);

	if (num_on_this_line > 13)
	{
	    printf("\n");
	    num_on_this_line = 0;
	}

	nwos_add_to_reference_list(&ref, &common_header.id);
    }

    if (num_on_this_line > 0) printf("\n");

    check_newline(line);
}
    

void import_class(char* name)
{
    char filename[36];
    char line[MAX_LINE_LENGTH];      /* this is large enough for all files except reference_lists */
    int i;
    FILE* fp = NULL;
    char* ref_list_line;

    for (i = 0; name[i] != '\0'; i++)
    {
	if (name[i] == ' ')
	{
	    filename[i] = '_';
	}
	else
	{
	  filename[i] = tolower(name[i]);
	}
    }

    filename[i++] = '.';
    filename[i++] = 'c';
    filename[i++] = 's';
    filename[i++] = 'v';
    filename[i] = '\0';

    fp = fopen(filename, "r");

    if (fp == NULL)
    {
	perror(filename);
	exit(1);
    }

    current_filename = filename;

    if (strcmp(name, "Class Definition") == 0)
    {
	while (fgets(line, sizeof(line), fp) != NULL) import_class_definition(line);
    }
    else if (strcmp(name, "Root") == 0)
    {
	while (fgets(line, sizeof(line), fp) != NULL) import_root(line);
    }
    else if (strcmp(name, "Spelling") == 0)
    {
	while (fgets(line, sizeof(line), fp) != NULL) import_spelling(line);
    }
    else if (strcmp(name, "Name") == 0)
    {
	while (fgets(line, sizeof(line), fp) != NULL) import_name(line);
    }
    else if (strcmp(name, "Word") == 0)
    {
	while (fgets(line, sizeof(line), fp) != NULL) import_word(line);
    }
    else if (strcmp(name, "Language") == 0)
    {
	while (fgets(line, sizeof(line), fp) != NULL) import_language(line);
    }
    else if (strcmp(name, "Us State") == 0)
    {
	while (fgets(line, sizeof(line), fp) != NULL) import_us_state(line);
    }
    else if (strcmp(name, "Us City") == 0)
    {
	while (fgets(line, sizeof(line), fp) != NULL) import_us_city(line);
    }
    else if (strcmp(name, "Area Code") == 0)
    {
	while (fgets(line, sizeof(line), fp) != NULL) import_area_code(line);
    }
    else if (strcmp(name, "Phone Number") == 0)
    {
	while (fgets(line, sizeof(line), fp) != NULL) import_phone_number(line);
    }
    else if (strcmp(name, "Mobile Phone") == 0)
    {
	while (fgets(line, sizeof(line), fp) != NULL) import_mobile_phone(line);
    }
    else if (strcmp(name, "Year") == 0)
    {
	while (fgets(line, sizeof(line), fp) != NULL) import_year(line);
    }
    else if (strcmp(name, "Month") == 0)
    {
	while (fgets(line, sizeof(line), fp) != NULL) import_month(line);
    }
    else if (strcmp(name, "Date") == 0)
    {
	while (fgets(line, sizeof(line), fp) != NULL) import_date(line);
    }
    else if (strcmp(name, "Gender") == 0)
    {
	while (fgets(line, sizeof(line), fp) != NULL) import_gender(line);
    }
    else if (strcmp(name, "Person") == 0)
    {
	while (fgets(line, sizeof(line), fp) != NULL) import_person(line);
    }
    else if (strcmp(name, "Reference List") == 0)
    {
	ref_list_line = malloc(MAX_REF_LIST_LINE_LENGTH);
	assert(ref_list_line != NULL);
	while (fgets(ref_list_line, MAX_REF_LIST_LINE_LENGTH, fp) != NULL) import_reference_list(ref_list_line);
	free(ref_list_line);
    }
}


int main(int argc, char* argv[])
{
    uint8 big_key[16 + 8 + 4];
    uint8 bf_key[16];
    uint32 linear;
    uint32 serial;
    int obj_fd;


    printf("\n");

    nwos_get_key_from_password(big_key, sizeof(big_key));

    printf("\n");

    memcpy(bf_key, big_key, 16);
    linear = ((uint32)big_key[16] << 24) | ((uint32)big_key[17] << 16) | ((uint32)big_key[18] << 8) | (uint32)big_key[19];
    memcpy(nwos_root_object_ref.id, big_key+20, 4);
    serial = ((uint32)big_key[24] << 24) | ((uint32)big_key[25] << 16) | ((uint32)big_key[26] << 8) | (uint32)big_key[27];

    /* create the file */

    obj_fd = open(DEFAULT_FILE, O_WRONLY | O_CREAT);    /* don't need large file to create it (I think) */

    if (obj_fd < 0)
    {
	perror(DEFAULT_FILE);
	exit(1);
    }

    if (fchmod(obj_fd, 0600) != 0)
    {
	perror(DEFAULT_FILE);
	exit(1);
    }

    if (write(obj_fd, MAGIC_NUMBER VERSION_STRING, 8) != 8)     /* used version string 0010 in eiffel version */
    {
	perror(DEFAULT_FILE);
	exit(1);
    }

    if (close(obj_fd) != 0)
    {
	perror(DEFAULT_FILE);
	exit(1);
    }


    nwos_initialize_objectify(bf_key, linear, serial, DEFAULT_TYPE, DEFAULT_FILE);

    import_class("Spelling");
    import_class("Name");
    import_class("Class Definition");
    import_class("Word");
    import_class("Language");

    import_class("Us State");
    import_class("Us City");

    import_class("Area Code");
    import_class("Phone Number");
    import_class("Mobile Phone");

    import_class("Year");
    import_class("Month");
    import_class("Date");

    import_class("Gender");
    import_class("Person");

    import_class("Reference List");

    import_class("Root");   /* this has to be last because it depends upon everything else existing */

    nwos_terminate_objectify();

    return 0;
}

