/*
--             This file is part of the New World OS project
--                 Copyright (C) 2007-2008  QRW Software
--           J. Scott Edwards - j.scott.edwards.nwos@gmail.com 
--                      http://www.qrwsoftware.com
--                      http://nwos.sourceforge.com
--
--   This program is free software: you can redistribute it and/or modify
--   it under the terms of the GNU General Public License as published by
--   the Free Software Foundation, either version 3 of the License, or
--   (at your option) any later version.
--
--   This program is distributed in the hope that it will be useful,
--   but WITHOUT ANY WARRANTY; without even the implied warranty of
--   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--   GNU General Public License for more details.
--
--   You should have received a copy of the GNU General Public License
--   along with this program, in the file LICENSE.  If not, see 
--   <http://www.gnu.org/licenses/>.
--
--   You can also contact me via paper mail at:
--
--      QRW Software
--      P.O. Box 27511
--      Salt Lake City, UT 84127-0511, USA.
--
--
-- $Log: combine_backup_files.c,v $
-- Revision 1.6  2008/09/01 00:10:49  jsedwards
-- Fix copyright year.  NO code changes.
--
-- Revision 1.5  2008/05/19 03:53:29  jsedwards
-- Increased the MAX_FILES to 100 from 64.
--
-- Revision 1.4  2007/11/18 13:25:45  jsedwards
-- Fix error test of return value from read_next_file_chunk.
--
-- Revision 1.3  2007/11/18 13:20:19  jsedwards
-- Fixed to handle file that is exactly MAX_BUFFER_SIZE. (still untestd)
--
-- Revision 1.2  2007/11/18 12:59:29  jsedwards
-- Added code to deal with files larger than MAX_BUFFER_SIZE. (Untested)
--
-- Revision 1.1  2007/11/17 21:42:55  jsedwards
-- Initial version of program to combine multiple backup (.bak) files into one
-- diff file.
--
*/

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "objectify_private.h"


#define MAX_FILES 100
#define MAX_BUFFER_SIZE 1048576
#define MAX_BLOCKS (MAX_BUFFER_SIZE / FILE_BLOCK_SIZE)



typedef struct {
    uint8 block[4];
    ObjRef ref;
    uint8 body[FILE_BLOCK_SIZE - sizeof(ObjRef) - 4];
} Dummy_Block;


typedef struct
{
    char* path;

    Disk_Header header1;
    uint8 unused1[FILE_BLOCK_SIZE - sizeof(Disk_Header)];

    Disk_Header header2;
    uint8 unused2[FILE_BLOCK_SIZE - sizeof(Disk_Header)];

    uint32 offset;

    int num_blocks;
    int index;

    Dummy_Block blocks[0];
} Backup_File;


/* Returns -1 if error, 0 if end of file, and > 0 if new chunk read in */

int read_next_file_chunk(Backup_File* file_ptr)
{
    FILE* fp;
    int block;
    int errsv;

    file_ptr->index = 0;

    if (file_ptr->num_blocks < MAX_BLOCKS)
    {
	file_ptr->num_blocks = 0;    /* end of file */
	return 0;
    }

    fp = fopen(file_ptr->path, "r");

    if (fp == NULL)
    {
	return -1;
    }

    file_ptr->offset += MAX_BLOCKS;

    if (fseek(fp, (off_t)file_ptr->offset * FILE_BLOCK_SIZE, SEEK_SET) < 0)
    {
	return -1;
    }

    block = 0;

    while (block < MAX_BLOCKS)
    {
	if (fread(&file_ptr->blocks[block].block, FILE_BLOCK_SIZE, 1, fp) != 1)
	{
	    if (feof(fp)) break;

	    errsv = errno;
	    fclose(fp);
	    errno = errsv;

	    return -1;
	}

	block++;
    }

    file_ptr->num_blocks = block;

    if (fclose(fp) != 0)
    {
	return -1;
    }

    return block;
}


int main(int argc, char* argv[])
{
    int i;
    int block;
    int num_input_files;
    int lowest;
    FILE* fp;
    uint32 ref1;
    uint32 ref2;

    Backup_File* files[MAX_FILES];

    for (i = 0; i < MAX_FILES; i++) files[i] = NULL;

    if (argc > MAX_FILES + 2)
    {
	fprintf(stderr, "This version can only combine a maximum of %d files.\n", MAX_FILES);
	exit(1);
    }

    fp = fopen(argv[argc-1], "r");

    if (fp != NULL)
    {
	printf("ERROR: output file: %s already exists, will not overwrite.\n", argv[argc-1]);
	fclose(fp);
	exit(1);
    }

    num_input_files = argc - 2;  /* ignore program name and output file name */

    for (i = 0; i < num_input_files; i++)    /* the last name is the output file */
    {
	printf("%s: ", argv[i+1]);
	fflush(stdout);

	files[i] = malloc(sizeof(Backup_File) + MAX_BUFFER_SIZE);   /* 1 MB for blocks */

	if (files[i] == NULL)
	{
	    perror("allocating buffer");
	    fclose(fp);
	    exit(1);
	}

	files[i]->path = argv[i+1];
	files[i]->index = 0;

	fp = fopen(files[i]->path, "r");

	if (fp == NULL)
	{
	    perror(NULL);
	    exit(1);
	}

	if (fread(&files[i]->header1, FILE_BLOCK_SIZE, 1, fp) != 1)
	{
	    perror(NULL);
	    fclose(fp);
	    exit(1);
	}

	if (i > 0)  /* check to make sure the headers match up */
	{
	    if (memcmp(&files[i-1]->header2, &files[i]->header1, sizeof(Disk_Header)) != 0)
	    {
		fprintf(stderr, "ERROR: header 1 doesn't match file '%s' header 2\n", files[i-1]->path);
		exit(1);
	    }
	}

	if (fread(&files[i]->header2, FILE_BLOCK_SIZE, 1, fp) != 1)
	{
	    perror(NULL);
	    fclose(fp);
	    exit(1);
	}

        block = 0;

	files[i]->offset = 2;

	while (block < MAX_BLOCKS)
	{
	    if (fread(&files[i]->blocks[block].block, FILE_BLOCK_SIZE, 1, fp) != 1)
	    {
		if (feof(fp)) break;

		perror(NULL);
		fclose(fp);
		exit(1);
	    }

	    block++;
	}

	files[i]->num_blocks = block;

	if (fclose(fp) != 0)
	{
	    perror(NULL);
	    exit(1);
	}

	printf("%d blocks\n", files[i]->num_blocks);

	if (files[i]->num_blocks < MAX_BLOCKS)  /* reduce the memory used to what is actually used */
	{
	    files[i] = realloc(files[i], sizeof(Backup_File) + files[i]->num_blocks * FILE_BLOCK_SIZE);
	}

	assert(files[i] != NULL);  /* when reducing memory could it ever fail? */
    }


    /* now we've read all the input files, write the output file */

    fp = fopen(argv[argc-1], "w");

    if (fp == NULL)
    {
	perror(argv[argc-1]);
	exit(1);
    }

    if (fwrite(&files[0]->header1, FILE_BLOCK_SIZE, 1, fp) != 1)
    {
	perror(NULL);
	fclose(fp);
	exit(1);
    }

    if (fwrite(&files[num_input_files-1]->header2, FILE_BLOCK_SIZE, 1, fp) != 1)
    {
	perror(NULL);
	fclose(fp);
	exit(1);
    }

    i = 0;
    while (i < num_input_files)
    {
	for (lowest = 0; lowest < num_input_files; lowest++) if (files[lowest] != NULL) break;

	for (i = lowest + 1; i < num_input_files; i++)
	{
	    if (files[i] != NULL)
	    {
		ref1 = nwos_ref_to_word(&files[lowest]->blocks[files[lowest]->index].ref);
		ref2 = nwos_ref_to_word(&files[i]->blocks[files[i]->index].ref);

		if (ref1 == ref2)  /* second supercedes first */
		{
		    files[lowest]->index++;   /* skip over this block, won't be used */

		    if (files[lowest]->index == files[lowest]->num_blocks)   /* finished with this file? */
		    {
			if (read_next_file_chunk(files[lowest]) < 0)
			{
			    perror(files[lowest]->path);
			    fclose(fp);
			    exit(1);
			}
			if (files[lowest]->num_blocks == 0)             /* yes */
			{
			    free(files[lowest]);
			    files[lowest] = NULL;
			}
		    }

		    lowest = i;
		}
		else if (ref1 > ref2)
		{
		    lowest = i;
		}
	    }
	}

	if (fwrite(&files[lowest]->blocks[files[lowest]->index].block, FILE_BLOCK_SIZE, 1, fp) != 1)
	{
	    perror(argv[argc-1]);
	    fclose(fp);
	    exit(1);
	}

	files[lowest]->index++;   /* this block is done */

	if (files[lowest]->index == files[lowest]->num_blocks)   /* finished with this file */
	{
	    if (read_next_file_chunk(files[lowest]) < 0)
	    {
		perror(files[lowest]->path);
		fclose(fp);
		exit(1);
	    }
	    if (files[lowest]->num_blocks == 0)             /* yes */
	    {
		free(files[lowest]);
		files[lowest] = NULL;
	    }
	}

	/* set i to the first file still in play, so we know when we are done */
	for (i = 0; i < num_input_files; i++) if (files[i] != NULL) break;
    }

    if (fclose(fp) != 0)
    {
	perror(NULL);
	exit(1);
    }

    return 0;
}

