/*
--             This file is part of the New World OS project
--                 Copyright (C) 2006-2008  QRW Software
--           J. Scott Edwards - j.scott.edwards.nwos@gmail.com 
--                      http://www.qrwsoftware.com
--                      http://nwos.sourceforge.com
--
--   This program is free software: you can redistribute it and/or modify
--   it under the terms of the GNU General Public License as published by
--   the Free Software Foundation, either version 3 of the License, or
--   (at your option) any later version.
--
--   This program is distributed in the hope that it will be useful,
--   but WITHOUT ANY WARRANTY; without even the implied warranty of
--   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--   GNU General Public License for more details.
--
--   You should have received a copy of the GNU General Public License
--   along with this program, in the file LICENSE.  If not, see 
--   <http://www.gnu.org/licenses/>.
--
--   You can also contact me via paper mail at:
--
--      QRW Software
--      P.O. Box 27511
--      Salt Lake City, UT 84127-0511, USA.
--
--
-- $Log: import_file.c,v $
-- Revision 1.40  2008/09/01 18:53:48  jsedwards
-- Fixed Bug #2043130 - asserts if given an absolute path.
--
-- Revision 1.39  2008/09/01 03:13:42  jsedwards
-- Change for new nwos_initialize_objectify calling convention (doesn't pass
-- back root_object_reference anymore) and removed call to nwos_set_root_object
-- because initialize calls it now.
--
-- Revision 1.38  2008/08/30 12:46:12  jsedwards
-- Removed code and variables to read pass phrase and pass it to initialize,
-- and change parameters passed to initialize.
--
-- Revision 1.37  2008/07/13 14:18:58  jsedwards
-- Feature Request #1856369 - Added handling of new IMPORTED_DATA result from
-- nwos_create_file.
--
-- Revision 1.36  2008/07/11 13:01:05  jsedwards
-- Added include of limits.h to define PATH_MAX.
--
-- Revision 1.35  2008/06/23 13:40:18  jsedwards
-- Fixed to remove the leading "./" from the path if doing "-r .".
--
-- Revision 1.34  2008/06/23 04:31:40  jsedwards
-- The nwos_normalize_path function changed to return void instead of book,
-- accept "." and ".." paths so calls were changed to deal with it.
--
-- Revision 1.33  2008/06/22 19:01:18  jsedwards
-- Reversed order of expressions in if statement so that it tests the size
-- before testing the array (should not have caused a problem anyway).
--
-- Revision 1.32  2008/06/21 16:17:42  jsedwards
-- Moved cleanup_path function to file.c, renamed to nwos_normalize_path, and
-- added size parameter.
--
-- Revision 1.31  2008/06/21 03:49:20  jsedwards
-- Fixed cleanup_path to do multiple slashes, so it will eliminate '///' or
-- '////' as well as '//'.
--
-- Revision 1.30  2008/06/21 03:44:18  jsedwards
-- Added cleanup_path function and fixed to eliminate duplicate slashes in
-- path.
--
-- Revision 1.29  2008/06/21 02:52:32  jsedwards
-- Fixed spacing on line 261.  NO code changes.
--
-- Revision 1.28  2008/06/21 02:12:14  jsedwards
-- Added print statement left out in previous check in.
--
-- Revision 1.27  2008/06/19 13:20:24  jsedwards
-- Changed to allow multiple paths to files when importing recursively.
--
-- Revision 1.26  2008/06/19 11:50:51  jsedwards
-- Added asserts to make sure we don't create any paths with a double slash.
-- NOTE: have NOT fixed the code yet to eliminate the double paths so if you
-- do import_file -r xyzzy/ it will assert.
--
-- Revision 1.25  2008/06/12 03:42:25  jsedwards
-- Moved 'check_invalid_type' and 'check_directory_is_all_ok' functions to
-- file.c and added 'nwos_' prefix.
--
-- Revision 1.24  2008/04/28 12:50:32  jsedwards
-- Added initializer for 'result' in process_file function so that the compiler
-- won't complain about it being uninitialized.
--
-- Revision 1.23  2008/04/26 13:00:45  jsedwards
-- Removed --add-revision code from main() and put it in process_file() so now
-- it can add-revision when recursively processing directories.
--
-- Revision 1.22  2008/04/26 12:01:00  jsedwards
-- Took code from main() and created new process_file function that can
-- recursively do directories.
--
-- Revision 1.21  2008/04/26 00:26:57  jsedwards
-- Added code to deal with recursively doing directories (Incomplete!).
--
-- Revision 1.20  2008/02/03 01:00:35  jsedwards
-- Changed DEFAULT_TYPE_RW to READ_WRITE.
--
-- Revision 1.19  2007/11/04 13:06:02  jsedwards
-- Added an extra note to look at Security_Levels.txt file in display_usage.
--
-- Revision 1.18  2007/11/04 13:00:43  jsedwards
-- Added display usage.
--
-- Revision 1.17  2007/10/08 12:40:52  jsedwards
-- Added call to restart ID generation for each file.
--
-- Revision 1.16  2007/10/07 03:56:09  jsedwards
-- Renamed 'nwos_set_block_estimate' to 'nwos_check_blocks_available' and
-- changed to skip the file if not enough space available.
--
-- Revision 1.15  2007/09/02 19:45:31  jsedwards
-- Added calls to set the block estimate.
--
-- Revision 1.14  2007/09/01 21:52:57  jsedwards
-- Changed 'encryption_options' table to 'security_options' to match up with
-- changes made earlier (Aug 12th).
--
-- Revision 1.13  2007/08/12 20:45:00  jsedwards
-- Change all of the "Encryption Level" stuff to "Security Level" because it
-- doesn't really change the encryption at all, all it does is change the
-- randomization of where objects are stored.
--
-- Revision 1.12  2007/08/06 02:22:45  jsedwards
-- Added "high" security option.
--
-- Revision 1.11  2007/07/24 04:47:03  jsedwards
-- Added "--add-revision" option and code to add new revision to a file that
-- is already in the system.
--
-- Revision 1.10  2007/07/01 19:44:12  jsedwards
-- Upgrade to GPLv3.
--
-- Revision 1.9  2007/02/11 16:58:26  jsedwards
-- Changed so DEFAULT_TYPE has to specify RO (Read-Only) or RW (Read-Write).
--
-- Revision 1.8  2007/01/27 14:59:03  jsedwards
-- Changed to handle new possible return values (duplicate path or file).
--
-- Revision 1.7  2007/01/17 13:17:43  jsedwards
-- Added security level options.
--
-- Revision 1.6  2007/01/07 20:31:55  jsedwards
-- Added call to log arguments.
--
-- Revision 1.5  2007/01/07 03:27:06  jsedwards
-- Added call to set encryption level (Low by default).
--
-- Revision 1.4  2006/12/07 14:10:25  jsedwards
-- Removed call to file_setup, no longer needed.
--
-- Revision 1.3  2006/11/11 12:01:03  jsedwards
-- Update e-mail address to something that works.
--
-- Revision 1.2  2006/10/26 01:51:27  jsedwards
-- Merged alpha_05_branch back into main trunk.
--
-- Revision 1.1.2.4  2006/10/13 12:58:52  jsedwards
-- Added printing of time to import each file.
--
-- Revision 1.1.2.3  2006/10/13 11:48:27  jsedwards
-- Add call to flush block maps after each file is read.
--
-- Revision 1.1.2.2  2006/10/03 12:53:07  jsedwards
-- Changed so that instead of calling a separate routine after initialize to
-- change the already opened storage, you call it now with a type of storage
-- parameter and a path to the storage.  The problem with the other way was
-- if you tried reading a compressed file on another machine it tried to open
-- the default file which didn't exist.
--
-- Revision 1.1.2.1  2006/09/18 01:45:37  jsedwards
-- Program to store a file into objectify.
--
*/

#include <assert.h>
#include <dirent.h>
#include <limits.h>    /* define PATH_MAX */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <time.h>
#include "objectify.h"


struct options {
  Security_Level level;
  char* option;
  char* description;
} security_options[] = 
  {
    { Security_Minimal,  "--minimal-security",  "minimal"  },
    { Security_Very_Low, "--very-low-security", "very low" },
    { Security_Low,      "--low-security",      "low"      },
    { Security_Medium,   "--medium-security",   "medium"   },
    { Security_High,     "--high-security",     "high"   },
    { Security_Extreme,  "--extreme-security",  "extreme"  },
  };

#define NUM_OPTIONS (sizeof(security_options)/sizeof(struct options))


static int result_counts[NUM_OBJ_RESULTS + 1];   /* allow one extra space for files that wouldn't fit */


static void display_usage(char* prog_name)
{
    int i;

    fprintf(stderr, "usage: %s [option] file [option] file...\n", prog_name);
    fprintf(stderr, "   or: %s --help\n", prog_name);
    fprintf(stderr, "\n");
    fprintf(stderr, "  --add-revision\tlink this file to a previous version.\n");
    fprintf(stderr, "  --help\t\tdisplay this message.\n");
    fprintf(stderr, "  -r, --recursive\trecurse through directories.\n");

    for (i = 0; i < NUM_OPTIONS; i++)
    {
	fprintf(stderr, "  %s\tuse %s security level for following files.\n",
		security_options[i].option, security_options[i].description);
    }

    fprintf(stderr, "\n");

    fprintf(stderr, "See the file doc/Security_Levels.txt for more info on the security levels.\n");

    fprintf(stderr, "\n");
}


typedef enum { Do_Not_Add_Revision, Add_Revision, Recursive_Add_Revision} Add_Revision_Type;

void process_file(const char* path, Add_Revision_Type add_revision, bool do_recursive)
{
    struct stat stat_struct;
    DIR* dp;
    struct dirent *dir_entry;
    const char* p;
    char full_path[PATH_MAX];
    int i;
    time_t start_time;
    ObjCreateResult result = ERROR_OCCURRED;
    ObjRef ref;
    ObjRef path_ref;
    bool Unknown_result_in_switch_statement = false;

    assert(strstr(path, "//") == NULL);

    if (lstat(path, &stat_struct) != 0)
    {
	perror(path);
	result_counts[ERROR_OCCURRED]++;
	return;
    }

    if (S_ISDIR(stat_struct.st_mode))
    {
	dp = opendir(path);
	if (dp == NULL)
	{
	    perror(path);
	    result_counts[ERROR_OCCURRED]++;
	    return;
	}

	dir_entry = readdir(dp);
	while (dir_entry != NULL)
	{
	    if (strcmp(dir_entry->d_name, ".") != 0 && strcmp(dir_entry->d_name, "..") != 0)
	    {
		// Construct the full path name

		i = 0;

		if (strcmp(path, ".") != 0)
		{
		    for (p = path; *p != '\0' && i < PATH_MAX; p++) full_path[i++] = *p;

		    if (i < PATH_MAX && full_path[i-1] != '/') full_path[i++] = '/';
		}

		for (p = dir_entry->d_name; *p != '\0' && i < PATH_MAX; p++) full_path[i++] = *p;

		if (i < PATH_MAX)
		{
		    full_path[i] = '\0';
		}
		else
		{
		    fprintf(stderr, "ERROR: path %s/%s exceeds maximum size of a path (PATH_MAX = %d)\n",
			    path, dir_entry->d_name, PATH_MAX);
		    result_counts[ERROR_OCCURRED]++;
		    return;
		}

		assert(strstr(full_path, "//") == NULL);

		if (add_revision == Do_Not_Add_Revision)
		{
		    process_file(full_path, Do_Not_Add_Revision, do_recursive);
		}
		else
		{
		    process_file(full_path, Recursive_Add_Revision, do_recursive);
		}
	    }

	    dir_entry = readdir(dp);
	}

	if (closedir(dp) != 0)
	{
	    perror(path);
	    result_counts[ERROR_OCCURRED]++;
	}
    }    
    else    // must be a regular file
    {
	assert(S_ISREG(stat_struct.st_mode));

	if (nwos_check_blocks_available(nwos_estimate_blocks_for_file(path)))
	{
	    start_time = time(NULL);

	    switch (add_revision)
	    {
	      case Do_Not_Add_Revision:
		result = nwos_create_file(path, &ref, do_recursive); /* allow multiple paths for the same file if recursive */
		break;

	      case Add_Revision:
		result = nwos_add_new_revision_of_file(path, &ref);
		break;

	      case Recursive_Add_Revision:
		if (nwos_find_file_path(path, &path_ref))   /* file with the same name exists */
		{
		    result = nwos_add_new_revision_of_file(path, &ref);
		}
		else
		{
		    result = nwos_create_file(path, &ref, true); /* no file with the same name exists, allow multiple paths */
		}
		break;
	    }

	    result_counts[result]++;

	    switch(result)
	    {
	      case CREATED_NEW:
		printf("file created: %02x%02x%02x%02x   time: %d seconds\n", 
		       ref.id[0], ref.id[1], ref.id[2], ref.id[3], 
		       (int) (time(NULL) - start_time));
		break;

	      case IMPORTED_DATA:
		printf("imported data: %02x%02x%02x%02x\n", ref.id[0], ref.id[1], ref.id[2], ref.id[3]);
		break;

	      case CREATED_NEW_REVISION:
		printf("file created: %02x%02x%02x%02x   time: %d seconds\n", 
		       ref.id[0], ref.id[1], ref.id[2], ref.id[3], 
		       (int) (time(NULL) - start_time));
		break;

	      case FOUND_EXISTING:
		printf("existing file found: %02x%02x%02x%02x\n", ref.id[0], ref.id[1], ref.id[2], ref.id[3]);
		break;

	      case DUPLICATE_PATH:
		printf("WARNING: a file by the name of: %s already exists and --add-revision.\n", path);
		printf("         was not specified, it will be skipped!!\n");
		break;

	      case PATH_NOT_FOUND:
		fprintf(stderr, "Error: no previous version of '%s' exists, will be skipped\n", path);
		break;

	      case DUPLICATE_FILE:
		printf("WARNING: a file with the same size, md5, and sha1 sums already exists.\n");
		printf("         In this version it will be skipped!!\n");
		break;

	      case MULTIPLE_FILES:
		fprintf(stderr, "Error: multiple files with the name %s exist, this version", path);
		fprintf(stderr, "cannot cope with them.\n");
		break;

	      case CREATED_NEW_PATH_FOR_EXISTING_FILE:
		printf("Created new path: %s for existing file: %02x%02x%02x%02x\n",
		       path, ref.id[0], ref.id[1], ref.id[2], ref.id[3]);
		break;

	      case ERROR_OCCURRED:
		break;

	      default:   /* keep the compiler form complaining */
		assert(Unknown_result_in_switch_statement);
		break;
	    }

	    nwos_flush_bit_maps();
	}
	else  /* not enough space for this file */
	{
	    fprintf(stderr, "Error: not enough space left for this file, skipped\n");
	    result_counts[NUM_OBJ_RESULTS]++;   /* use last slot to count too full errors */
	}
    }
}


int main(int argc, char* argv[])
{
    struct stat stat_struct;
    int i;
    int j;
    char path[PATH_MAX];
    bool do_recursive = false;
    const char* invalid_type;

    for (i = 0; i < (NUM_OBJ_RESULTS + 1); i++) result_counts[i] = 0;

    printf("\n");

    if (argc < 2)
    {
	display_usage(argv[0]);
	exit(1);
    }

    for (i = 1; i < argc; i++)
    {
	if (argv[i][0] == '-')
	{
	    for (j = 0; j < NUM_OPTIONS; j++)
	    {
		if (strcmp(argv[i], security_options[j].option) == 0) break;
	    }

	    if (j == NUM_OPTIONS)
	    {
		if (strcmp(argv[i], "--add-revision") == 0)
		{
		    i++;

		    if (i == argc || argv[i][0] == '-')
		    {
			fprintf(stderr, "Error: a file must be specified after --add-revision\n");
			exit(1);
		    }
		}
		else if (strcmp(argv[i], "-r") == 0 || strcmp(argv[i], "--recursive") == 0)
		{
		    do_recursive = true;
		}
		else if (strcmp(argv[i], "--help") == 0)
		{
		    display_usage(argv[0]);
		    exit(1);
		}
		else
		{
		    fprintf(stderr, "Unknown option: %s\n", argv[i]);
		    display_usage(argv[0]);
		    exit(1);
		}
	    }
	}
	else
	{
	    if (lstat(argv[i], &stat_struct) != 0)
	    {
		perror(argv[i]);
		exit(1);
	    }

	    invalid_type = nwos_check_invalid_type(stat_struct.st_mode);

	    if (invalid_type != NULL)
	    {
		fprintf(stderr, "ERROR: %s is a %s.\n", argv[i], invalid_type);
	    }

	    if (S_ISDIR(stat_struct.st_mode))    // it's not a normal file, investigate further
	    {
		if (do_recursive)
		{
		    nwos_check_directory_is_all_ok(argv[i]);
		}
		else
		{
		    fprintf(stderr, "ERROR: %s is a directory and -r or --recursive was not specified.\n", argv[i]);
		    exit(1);
		}
	    }    
	    else if (!S_ISREG(stat_struct.st_mode))
	    {
		fprintf(stderr, "ERROR: %s is unrecognized type of file.\n", argv[i]);
		exit(1);
	    }
	}
    }


    nwos_log_arguments(argc, argv);

    nwos_initialize_objectify(READ_WRITE, DEFAULT_FILE);

    nwos_set_security_level(Security_Low);


    for (i = 1; i < argc; i++)
    {
	nwos_restart_id_generation();   /* make sure each file starts with it's own sequence of id's */

	if (argv[i][0] == '-')
	{
	    if (strcmp(argv[i], "--add-revision") == 0)
	    {
		i++;    /* point to the file argument */

		assert(i < argc && argv[i][0] != '-');

		if (strlen(argv[i]) + 1 > sizeof(path))
		{
		    fprintf(stderr, "WARNING: the path '%s' is too long for the buffer (exceeds %d characters) - ignored.\n", argv[i], sizeof(path));
		}
		else if (*argv[i] == '/')
		{
		    fprintf(stderr, "WARNING: the path '%s' is absolute and this version cannot process it - ignored.\n", argv[i]);
		}
		else
		{
		    nwos_normalize_path(path, argv[i], sizeof(path));

		    if (path[0] == '.' && path[1] == '.' && path[2] == '/')
		    {
			fprintf(stderr, "ERROR: this version cannot import relative to the parent directory: %s\n", argv[i]);
			fprintf(stderr, "Will be skipped!\n");
		    }
		    else
		    {
			process_file(path, Add_Revision, do_recursive);
		    }
		}
	    }
	    else
	    {
		for (j = 0; j < NUM_OPTIONS; j++)
		{
		    if (strcmp(argv[i], security_options[j].option) == 0) break;
		}

		if (j < NUM_OPTIONS)
		{
		    printf("setting security level to: %s\n", security_options[j].description);

		    nwos_set_security_level(security_options[j].level);
		}
		else
		{
		    assert(strcmp(argv[i], "-r") == 0 || strcmp(argv[i], "--recursive") == 0);
		}
	    }
	}
	else
	{
	    if (strlen(argv[i]) + 1 > sizeof(path))
	    {
		fprintf(stderr, "WARNING: the path '%s' is too long for the buffer (exceeds %d characters) - ignored.\n", argv[i], sizeof(path));
	    }
	    else if (*argv[i] == '/')
	    {
		fprintf(stderr, "WARNING: the path '%s' is absolute and this version cannot process it - ignored.\n", argv[i]);
	    }
	    else
	    {
		nwos_normalize_path(path, argv[i], sizeof(path));

		if (path[0] == '.' && path[1] == '.' && path[2] == '/')
		{
		    fprintf(stderr, "ERROR: this version cannot import relative to the parent directory: %s\n", argv[i]);
		    fprintf(stderr, "Will be skipped!\n");
		}
		else
		{
		    process_file(path, Do_Not_Add_Revision, do_recursive);
		}
	    }
	}
    }

    printf("files created:   %d\n", result_counts[CREATED_NEW]);
    printf("imported data:   %d\n", result_counts[IMPORTED_DATA]);
    printf("new revisions:   %d\n", result_counts[CREATED_NEW_REVISION]);
    printf("new paths:       %d\n", result_counts[CREATED_NEW_PATH_FOR_EXISTING_FILE]);
    printf("existing files:  %d\n", result_counts[FOUND_EXISTING]);
    printf("path not found:  %d\n", result_counts[PATH_NOT_FOUND]);
    printf("duplicate paths: %d\n", result_counts[DUPLICATE_PATH]);
    printf("duplicate files: %d\n", result_counts[DUPLICATE_FILE]);
    printf("multiple files:  %d\n", result_counts[MULTIPLE_FILES]);
    printf("files w/errors:  %d\n", result_counts[ERROR_OCCURRED]);
    printf("too full to add: %d\n", result_counts[NUM_OBJ_RESULTS]);

    nwos_terminate_objectify();

    return 0;
}


