/*
--             This file is part of the New World OS project
--                 Copyright (C) 2005-2008  QRW Software
--           J. Scott Edwards - j.scott.edwards.nwos@gmail.com 
--                      http://www.qrwsoftware.com
--                      http://nwos.sourceforge.com
--
--   This program is free software: you can redistribute it and/or modify
--   it under the terms of the GNU General Public License as published by
--   the Free Software Foundation, either version 3 of the License, or
--   (at your option) any later version.
--
--   This program is distributed in the hope that it will be useful,
--   but WITHOUT ANY WARRANTY; without even the implied warranty of
--   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--   GNU General Public License for more details.
--
--   You should have received a copy of the GNU General Public License
--   along with this program, in the file LICENSE.  If not, see 
--   <http://www.gnu.org/licenses/>.
--
--   You can also contact me via paper mail at:
--
--      QRW Software
--      P.O. Box 27511
--      Salt Lake City, UT 84127-0511, USA.
--
--
-- $Log: number.c,v $
-- Revision 1.10  2008/09/01 00:10:52  jsedwards
-- Fix copyright year.  NO code changes.
--
-- Revision 1.9  2008/08/31 21:53:53  jsedwards
-- Added an assert around calls to nwos_read_variable_sized_object_from_disk
-- and nwos_read_object_from_disk because now they return false when they fail
-- instead of asserting themselves.
--
-- Revision 1.8  2007/07/01 19:44:12  jsedwards
-- Upgrade to GPLv3.
--
-- Revision 1.7  2006/12/21 12:59:47  jsedwards
-- Hack to make it compile with new split public and private classes,
-- NON-FUNCTIONAL!
--
-- Revision 1.6  2006/12/10 19:28:36  jsedwards
-- Added code to skip over leading zeros in a number.  Also fixed cardinal
-- number to string routine to verify it is an cardinal number.
--
-- Revision 1.5  2006/12/10 15:21:34  jsedwards
-- Added cardinal and ordinal number to string functions.
--
-- Revision 1.4  2006/12/10 15:05:07  jsedwards
-- Added find and create ordinal number routines.
--
-- Revision 1.3  2006/12/10 12:05:41  jsedwards
-- Changed to deal with numbers as strings of ascii decimal digits instead of
-- a uint32 and increased max_digits to 127 for now.
--
-- Revision 1.2  2006/12/09 18:55:37  jsedwards
-- Fixed bug using uninitialized variable to compute size of cardinal object.
-- Also verify it is a cardinal object in find before reading it.
--
-- Revision 1.1  2006/12/09 16:17:14  jsedwards
-- Initial version with only Cardinal Number functions (untested).
--
*/

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "crc32.h"
#include "objectify_private.h"


#define MAX_DIGITS 127



static size_t get_cardinal_object_size(void* cardinal_obj)
{
    int count = ((C_struct_Cardinal_Number*)cardinal_obj)->count;

    assert(0 < count && count <= MAX_DIGITS);

    return sizeof(C_struct_Cardinal_Number) + count;
}


/* returns true if all characters in string are ascii decimal digits */
static bool is_number(char* string)
{
    char *p;

    if (*string == '\0') return false;  /* can't be an empty string */

    for (p = string; *p != '\0'; p++) if (!isdigit(*p)) break;

    return *p == '\0';
}


static char* skip_leading_zeros(char* number)
{
    char* result;

    assert(is_number(number));

    result = number;

    while (*result == '0') result++;

    if (*result == '\0') result--;

    return result;
}


bool nwos_find_cardinal_number(char* number, ObjRef* ref)
{
    C_struct_Class_Definition class_def_obj;
    uint8 kludge[FILE_BLOCK_SIZE];
    C_struct_Cardinal_Number* cardinal_obj_ptr = (C_struct_Cardinal_Number*)kludge;
    ObjRef cardinal_class_ref;
    ObjRef class_ref;
    ReferenceList* ref_list;
    int num_digits;
    int num_numbers;
    int i;
    char* ptr_first_digit;

    assert(is_number(number));

    ptr_first_digit = skip_leading_zeros(number);

    num_digits = strlen(ptr_first_digit);

    assert(num_digits <= MAX_DIGITS);


    assert(nwos_find_public_class_definition("CARDINAL NUMBER", &cardinal_class_ref));

    nwos_read_class_definition(&cardinal_class_ref, &class_def_obj);

    ref_list = nwos_malloc_reference_list(&class_def_obj.header.object.references);

    num_numbers = ref_list->common_header.num_refs;

    /* printf("num_months: %d\n", num_months); */

    for (i = 0; i < num_numbers; i++)
    {
	nwos_get_object_class(&ref_list->references[i], &class_ref);   /* find it's class */

	if (is_same_object(&cardinal_class_ref, &class_ref))
	{
	    assert(nwos_read_variable_sized_object_from_disk(&ref_list->references[i], kludge, sizeof(kludge), &get_cardinal_object_size));

	    if (cardinal_obj_ptr->count == num_digits)   /* this could be it */
	    {
		if (memcmp(cardinal_obj_ptr->digits, ptr_first_digit, num_digits) == 0)   /* found a match */
		{
		    memcpy(ref, &ref_list->references[i], sizeof(ObjRef));
		    break;
		}
	    }
	}
    }

    nwos_free_reference_list(ref_list);
    ref_list = NULL;

    return i < num_numbers;
}


/* Find existing cardinal_number or create new */

ObjCreateResult nwos_create_cardinal_number(char* number, ObjRef* ref)
{
    uint8 kludge[FILE_BLOCK_SIZE];
    C_struct_Cardinal_Number* ptr_cardinal_obj = (C_struct_Cardinal_Number*)kludge;
    ObjCreateResult result = FOUND_EXISTING;
    ObjRef class_ref;
    int num_digits;
    char* ptr_first_digit;

    assert(is_number(number));

    ptr_first_digit = skip_leading_zeros(number);

    num_digits = strlen(ptr_first_digit);

    assert(num_digits <= MAX_DIGITS);


    /* first find out if we already have this cardinal_number */

    if (!nwos_find_cardinal_number(ptr_first_digit, ref))   /* didn't find it */
    {
	assert(nwos_find_public_class_definition("CARDINAL NUMBER", &class_ref));

	memset(kludge, 0, sizeof(kludge));  /* zero it out */

	/* remember ptr_cardinal_obj points to the kludge buffer */

	nwos_generate_new_id(ref);

	nwos_fill_in_common_header(&ptr_cardinal_obj->header.common, ref, &class_ref);

	ptr_cardinal_obj->count = num_digits;

	memcpy(ptr_cardinal_obj->digits, ptr_first_digit, num_digits);

	nwos_create_reference_list(ref, &ptr_cardinal_obj->header.object.references);

	nwos_crc32_calculate((uint8*) &ptr_cardinal_obj->header.object, sizeof(ObjectHeader), ptr_cardinal_obj->header.common.header_chksum);

	nwos_crc32_calculate((uint8*) &ptr_cardinal_obj->count, get_cardinal_object_size(kludge) - sizeof(EveryObject), ptr_cardinal_obj->header.common.data_chksum);

	nwos_write_object_to_disk(ref, kludge, get_cardinal_object_size(kludge));

	nwos_add_to_references(ref, &class_ref);

	ptr_cardinal_obj = malloc(sizeof(C_struct_Cardinal_Number) + num_digits);
	assert(nwos_read_object_from_disk(ref, ptr_cardinal_obj, sizeof(C_struct_Cardinal_Number) + num_digits));
	assert(memcmp(kludge, ptr_cardinal_obj, sizeof(C_struct_Cardinal_Number) + num_digits) == 0);

	memset(kludge, 0, sizeof(kludge));  /* clear it */
	assert(nwos_read_variable_sized_object_from_disk(ref, kludge, sizeof(kludge), &get_cardinal_object_size));  /* read the other way */
	assert(memcmp(ptr_cardinal_obj, kludge, sizeof(C_struct_Cardinal_Number) + num_digits) == 0);

	free(ptr_cardinal_obj);
	ptr_cardinal_obj = NULL;

	result = CREATED_NEW;
    }

    return result;
}


/* returns true if the object passed in is a cardinal number */
static bool is_cardinal_number(ObjRef* ref)
{
    ObjRef cardinal_class_ref;
    ObjRef class_ref;

    assert(nwos_find_public_class_definition("CARDINAL NUMBER", &cardinal_class_ref));
    nwos_get_object_class(ref, &class_ref);   /* find it's class */

    return is_same_object(&cardinal_class_ref, &class_ref);
}


void nwos_cardinal_number_to_string(ObjRef* ref, char* string, size_t size)
{
    uint8 kludge[FILE_BLOCK_SIZE];
    C_struct_Cardinal_Number* cardinal_obj_ptr = (C_struct_Cardinal_Number*)kludge;
    int i;

    assert(is_cardinal_number(ref));

    assert(nwos_read_variable_sized_object_from_disk(ref, kludge, sizeof(kludge), &get_cardinal_object_size));
    assert(cardinal_obj_ptr->count < size);

    for (i = 0; i < cardinal_obj_ptr->count; i++) string[i] = cardinal_obj_ptr->digits[i];
    string[i] = '\0';
}



/*******************/
/* Ordinal numbers */
/*******************/

bool nwos_find_ordinal_number(ObjRef* cardinal_ref, ObjRef* ref)
{
    C_struct_Class_Definition class_def_obj;
    C_struct_Ordinal_Number ordinal_obj;
    ObjRef ordinal_class_ref;
    ObjRef class_ref;
    ReferenceList* ref_list;
    int num_numbers;
    int i;

    assert(!is_void_reference(cardinal_ref));
    assert(is_cardinal_number(cardinal_ref));

    assert(nwos_find_public_class_definition("ORDINAL NUMBER", &ordinal_class_ref));

    nwos_read_class_definition(&ordinal_class_ref, &class_def_obj);

    ref_list = nwos_malloc_reference_list(&class_def_obj.header.object.references);

    num_numbers = ref_list->common_header.num_refs;

    /* printf("num_months: %d\n", num_months); */

    for (i = 0; i < num_numbers; i++)
    {
	nwos_get_object_class(&ref_list->references[i], &class_ref);   /* find it's class */

	if (is_same_object(&ordinal_class_ref, &class_ref))
	{
	    assert(nwos_read_object_from_disk(&ref_list->references[i], &ordinal_obj, sizeof(ordinal_obj)));

	    if (is_same_object(&ordinal_obj.cardinal, cardinal_ref))   /* found a match */
	    {
		memcpy(ref, &ref_list->references[i], sizeof(ObjRef));
		break;
	    }
	}
    }

    nwos_free_reference_list(ref_list);
    ref_list = NULL;

    return i < num_numbers;
}


/* Find existing ordinal_number or create new */

ObjCreateResult nwos_create_ordinal_number(ObjRef* cardinal_ref, ObjRef* ref)
{
    C_struct_Ordinal_Number ordinal_obj;
    ObjCreateResult result = FOUND_EXISTING;
    ObjRef class_ref;

    assert(!is_void_reference(cardinal_ref));
    assert(is_cardinal_number(cardinal_ref));


    /* first find out if we already have this ordinal_number */

    if (!nwos_find_ordinal_number(cardinal_ref, ref))   /* didn't find it */
    {
	assert(nwos_find_public_class_definition("ORDINAL NUMBER", &class_ref));

	nwos_generate_new_id(ref);

	nwos_fill_in_common_header(&ordinal_obj.header.common, ref, &class_ref);

	copy_reference(&ordinal_obj.cardinal, cardinal_ref);

	nwos_create_reference_list(ref, &ordinal_obj.header.object.references);

	nwos_crc32_calculate((uint8*) &ordinal_obj.header.object, sizeof(ObjectHeader), ordinal_obj.header.common.header_chksum);
	nwos_crc32_calculate((uint8*) &ordinal_obj.cardinal, sizeof(ordinal_obj) - sizeof(EveryObject), ordinal_obj.header.common.data_chksum);

	nwos_write_object_to_disk(ref, &ordinal_obj, sizeof(ordinal_obj));

	{
	C_struct_Ordinal_Number* ptr_ordinal_obj;
	ptr_ordinal_obj = malloc(sizeof(C_struct_Ordinal_Number));
	assert(nwos_read_object_from_disk(ref, ptr_ordinal_obj, sizeof(C_struct_Ordinal_Number)));
	assert(memcmp(&ordinal_obj, ptr_ordinal_obj, sizeof(C_struct_Ordinal_Number)) == 0);
	free(ptr_ordinal_obj);
	ptr_ordinal_obj = NULL;
	}

	nwos_add_to_references(ref, &class_ref);
	nwos_add_to_references(ref, cardinal_ref);

	result = CREATED_NEW;
    }

    return result;
}


void nwos_ordinal_number_to_string(ObjRef* ref, char* string, size_t size)
{
    C_struct_Ordinal_Number ordinal_obj;

    assert(nwos_read_object_from_disk(ref, &ordinal_obj, sizeof(ordinal_obj)));

    nwos_cardinal_number_to_string(&ordinal_obj.cardinal, string, size);
}


